"""PKONE serial communicator."""
import asyncio
import re
from packaging import version

from mpf.platforms.base_serial_communicator import BaseSerialCommunicator
from mpf.platforms.pkone.pkone_extension import PKONEExtensionBoard
from mpf.platforms.pkone.pkone_lightshow import PKONELightshowBoard

MYPY = False
if MYPY:    # pragma: no cover
    from mpf.platforms.pkone.pkone import PKONEHardwarePlatform   # pylint: disable-msg=cyclic-import,unused-import

NANO_MIN_FW = '1.0'
EXTENSION_MIN_FW = '1.0'
LIGHTSHOW_MIN_FW = '1.0'


class PKONESerialCommunicator(BaseSerialCommunicator):

    """Handles the serial communication to the PKONE platform."""

    ignored_messages = ['PWD',  # Watchdog
                        ]

    __slots__ = ["part_msg", "send_queue", "remote_firmware", "remote_hardware_rev", "received_msg",
                 "max_messages_in_flight", "messages_in_flight", "send_ready"]

    # pylint: disable=too-many-arguments
    def __init__(self, platform: "PKONEHardwarePlatform", port, baud) -> None:
        """Initialize Serial Connection to PKONE Hardware.

        Args:
        ----
            platform(mpf.platforms.pkone.pkone.HardwarePlatform): the pkone hardware platform
            port: serial port
            baud: baud rate
        """
        self.send_queue = asyncio.Queue()
        self.remote_firmware = None
        self.remote_hardware_rev = None
        self.received_msg = b''
        self.max_messages_in_flight = 10
        self.messages_in_flight = 0

        self.send_ready = asyncio.Event()
        self.send_ready.set()

        super().__init__(platform, port, baud)

    async def _read_with_timeout(self, timeout):
        try:
            msg_raw = await asyncio.wait_for(self.readuntil(b'E'), timeout=timeout)
        except asyncio.TimeoutError:
            return ""
        return msg_raw.decode()

    async def _identify_connection(self):
        """Identify which controller this serial connection is talking to."""
        count = 0
        while True:
            if (count % 10) == 0:
                self.platform.debug_log("Sending 'PCN' command to port '%s'", self.port)

            count += 1
            self.writer.write('PCNE'.encode('ascii', 'replace'))
            msg = await self._read_with_timeout(.5)
            if msg.startswith('PCN'):
                break

            await asyncio.sleep(.5)

            if count == 100:
                raise AssertionError('No response from PKONE hardware on port {}'.format(self.port))

        # PCN (Determine connected controller board) reply is in the following format:
        # PCNF[Firmware rev]H[Hardware rev]E
        match = re.match('PCNF([0-9]+)H([0-9]+)E', msg)
        if not match:
            raise AssertionError(
                'Received an unexpected response. {} is not a recognized response to the PCN command.'.format(msg))

        self.remote_firmware = match[1][:-1] + '.' + match[1][-1]
        self.remote_hardware_rev = match[2]

        self.platform.log.info("Connected! "
                               "Board Type: PKONE Nano Controller, Firmware: %s, Hardware Rev: %s",
                               self.remote_firmware, self.remote_hardware_rev)

        self.machine.variables.set_machine_var("pkone_firmware", self.remote_firmware)
        '''machine_var: pkone_firmware

        desc: Holds the version number of the firmware for the Penny K Pinball PKONE controller that's connected.'''

        self.machine.variables.set_machine_var("pkone_hardware",
                                               "PKONE Nano Controller (rev {})".format(self.remote_hardware_rev))
        '''machine_var: pkone_hardware

        desc: Holds the model name and hardware revision number of the Penny K Pinball PKONE controller
        board that's connected.'''

        if version.parse(NANO_MIN_FW) > version.parse(self.remote_firmware):
            raise AssertionError('Firmware version mismatch. MPF requires '
                                 'the PKONE Nano Controller to be firmware {}, but yours is {}. '
                                 'Please update your firmware.'.
                                 format(NANO_MIN_FW, self.remote_firmware))

        # Reset the Nano controller and connected boards
        await self.reset_controller()

        # Determine what additional boards are connected to the Nano controller
        await self.query_pkone_boards()

        # Read the initial state of all switches
        await self.read_all_switches()

        self.platform.controller_connection = self

    async def reset_controller(self):
        """Reset the controller."""
        self.platform.debug_log('Resetting controller.')

        # this command returns several responses (one from each board, starting with the Nano controller)
        self.writer.write('PRSE'.encode())
        msg = ''
        while msg != 'PRSE' and not msg.startswith('PXX'):
            msg = (await self.readuntil(b'E')).decode()
            self.platform.debug_log("Got: %s", msg)

        if msg.startswith('PXX'):
            raise AssertionError('Received an error while resetting the controller: {}'.format(msg))

    async def query_pkone_boards(self):
        """Query the NANO processor to discover which additional boards are connected."""
        self.platform.debug_log('Querying PKONE boards...')

        # Determine connected add-on boards (PCB command)
        # Responses:
        # Extension board - PCB01XF11H1 = PCB[board number 0-7]XP[Y:48V, N: no 48V]F[firmware rev]H[hardware rev]
        # Lightshow board - PCB01LF10H1RGBW = PCB[board number 0-3]LF[firmware rev]H[hardware rev][firmware_type]
        # No board at the address: PCB[board number 0-7]N
        for address_id in range(8):
            self.writer.write('PCB{}E'.format(address_id).encode('ascii', 'replace'))
            msg = await self._read_with_timeout(.5)
            if msg == 'PCB{}NE'.format(address_id):
                self.platform.log.debug("No board at address ID {}".format(address_id))
                continue

            match = re.fullmatch('PCB([0-7])([XLN])F([0-9]+)H([0-9]+)(P[YN])?(RGB|RGBW)?E', msg)
            if not match:
                self.platform.log.warning("Received unexpected message from PKONE: {}".format(msg))

            if match.group(2) == "X":
                # Extension board
                firmware = match.group(3)[:-1] + '.' + match.group(3)[-1]
                hardware_rev = match.group(4)

                if version.parse(EXTENSION_MIN_FW) > version.parse(firmware):
                    raise AssertionError('Firmware version mismatch. MPF requires '
                                         'PKONE Extension boards to be at least firmware {}, but yours is {}. '
                                         'Please update your firmware.'.
                                         format(EXTENSION_MIN_FW, firmware))

                self.platform.debug_log('PKONE Extension Board {0}: '
                                        'Firmware: {1}, Hardware Rev: {2}'.format(address_id,
                                                                                  firmware, hardware_rev))

                self.platform.register_extension_board(PKONEExtensionBoard(address_id, firmware, hardware_rev))

            elif match.group(2) == "L":
                # Lightshow board
                firmware = match.group(3)[:-1] + '.' + match.group(3)[-1]
                hardware_rev = match.group(4)
                rgbw_firmware = match.group(6) == 'RGBW'

                if version.parse(LIGHTSHOW_MIN_FW) > version.parse(firmware):
                    raise AssertionError('Firmware version mismatch. MPF requires '
                                         'PKONE Lightshow boards to be at least firmware {}, but yours is {}. '
                                         'Please update your firmware.'.
                                         format(LIGHTSHOW_MIN_FW, firmware))

                self.platform.debug_log('PKONE Lightshow Board {0}: Firmware: {1} ({2}), '
                                        'Hardware Rev: {3}'.format(address_id,
                                                                   firmware,
                                                                   'RGBW' if rgbw_firmware else 'RGB',
                                                                   hardware_rev))

                self.platform.register_lightshow_board(PKONELightshowBoard(address_id,
                                                                           firmware,
                                                                           hardware_rev,
                                                                           rgbw_firmware))

            else:
                raise AttributeError("Unrecognized PKONE board type in message: {}".format(msg))

    async def read_all_switches(self):
        """Read the current state of all switches from the hardware."""
        self.platform.debug_log('Reading all switches.')
        for address_id in self.platform.pkone_extensions:
            self.writer.write('PSA{}E'.format(address_id).encode())
            msg = ''
            while not msg.startswith('PSA'):
                msg = (await self.readuntil(b'E')).decode()
                if not msg.startswith('PSA'):
                    self.platform.log.warning("Received unexpected message from PKONE: {}".format(msg))

            self.platform.process_received_message(msg)

    def _parse_msg(self, msg):
        self.received_msg += msg

        while True:
            pos = self.received_msg.find(b'E')

            # no more complete messages
            if pos == -1:
                break

            msg = self.received_msg[:pos]
            self.received_msg = self.received_msg[pos + 1:]

            self.messages_in_flight -= 1
            if self.messages_in_flight <= self.max_messages_in_flight or not self.read_task:
                self.send_ready.set()
            if self.messages_in_flight < 0:
                self.log.warning("Received more messages than were sent! Resetting!")
                self.messages_in_flight = 0

            if not msg:
                continue

            if msg.decode() not in self.ignored_messages:
                self.platform.process_received_message(msg.decode())

    def send(self, msg):
        """Send a message to the remote processor over the serial connection.

        Args:
        ----
            msg: Bytes of the message you want to send.
        """
        if self.debug:
            self.log.debug("%s sending: %s", self, msg)

        self.writer.write(msg.encode() + b'E')



"""Contains the MPF Plugin class."""

from mpf.core.logging import LogMixin


class MpfPlugin(LogMixin):

    """Base class for a plugin module."""

    __slots__ = ("machine", "name", "config")

    config_section = None

    def __init__(self, machine):
        """Initialize custom code but with custom logging formatter."""
        super().__init__()
        self.machine = machine
        self.name = type(self).__name__
        self.config = None

        if self.config_section and self.config_section not in self.machine.config:
            self.machine.log.debug('"%s:" section not found in machine '
                                   'configuration, so the %s will not be '
                                   'used.', self.config_section, self.name)

    def initialize(self):
        """Called when the plugin is enabled and loaded into MPF.

        Override with plugin-specific init behavior.
        """

    @property
    def is_plugin_enabled(self):
        """If false, this plugin will not be attached to the MPF process.

        Override with class-specific logic.
        """
        if self.config_section:
            return self.config_section in self.machine.config
        return True



"""Slide player for MPF to support GMC."""
from functools import partial

from mpf.config_players.plugin_player import PluginPlayer


class SlidePlayer(PluginPlayer):

    """Base class for part of the slide player which runs as part of MPF.

    Note: This class is loaded by MPF and everything in it is in the context of
    MPF, not the mc.
    """

    config_file_section = 'slide_player'
    show_section = 'slides'

    def _validate_config_item(self, device, device_settings):
        # device is slide name, device_settings is configuration

        device_settings = self.machine.config_validator.validate_config(
            'slide_player', device_settings)

        return_dict = dict()
        return_dict[device] = device_settings

        return return_dict

    def _register_trigger(self, event, **kwargs):
        """Register trigger via BCP for MC."""
        del kwargs
        client = self.machine.bcp.transport.get_named_client("local_display")
        if client:
            self.machine.bcp.interface.add_registered_trigger_event_for_client(client, event)
        else:
            self.machine.events.add_handler("bcp_clients_connected", partial(self._register_trigger, event))



"""Physical segment displays."""
from asyncio import Future
from typing import Optional, Dict, List

from mpf.core.clock import PeriodicTask
from mpf.core.rgb_color import RGBColor
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.placeholder_manager import TextTemplate
from mpf.core.system_wide_device import SystemWideDevice
from mpf.devices.segment_display.text_stack_entry import TextStackEntry
from mpf.devices.segment_display.transition_manager import TransitionManager
from mpf.devices.segment_display.transitions import TransitionRunner, TransitionBase
from mpf.devices.segment_display.segment_display_text import SegmentDisplayText, ColoredSegmentDisplayText
from mpf.platforms.interfaces.segment_display_platform_interface import FlashingType
from mpf.plugins.virtual_segment_display_connector import VirtualSegmentDisplayConnector

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.platforms.interfaces.segment_display_platform_interface import SegmentDisplayPlatformInterface     # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.core.platform import SegmentDisplayPlatform    # pylint: disable-msg=cyclic-import,unused-import; # noqa


class SegmentDisplayState:

    """Current State."""

    __slots__ = ["text", "flashing", "flash_mask"]

    def __init__(self, text: ColoredSegmentDisplayText, flashing: FlashingType, flash_mask: Optional[str] = None):
        """Class initializer."""
        self.text = text
        self.flashing = flashing
        self.flash_mask = flash_mask

    def __repr__(self):
        """Return str representation."""
        return '<TextStackEntry: {} (flashing: {} flashing_mask: {}) >'.format(
            self.text, self.flashing, self.flash_mask)

    def __eq__(self, other):
        """Compose two instances."""
        return self.text == other.text and self.flashing == other.flashing and \
            self.flash_mask == other.flash_mask


# pylint: disable=too-many-instance-attributes
@DeviceMonitor("text")
class SegmentDisplay(SystemWideDevice):

    """A physical segment display in a pinball machine."""

    __slots__ = ["hw_display", "size", "virtual_connector", "_text_stack", "_current_placeholder",
                 "_current_text_stack_entry", "_transition_update_task", "_current_transition", "_default_color",
                 "_current_state", "_current_placeholder_future"]

    config_section = 'segment_displays'
    collection = 'segment_displays'
    class_label = 'segment_display'

    def __init__(self, machine, name: str) -> None:
        """Initialize segment display device."""
        super().__init__(machine, name)
        self.hw_display = None                      # type: Optional[SegmentDisplayPlatformInterface]
        self.platform = None                        # type: Optional[SegmentDisplayPlatform]
        self.size = None                            # type: Optional[int]

        self.virtual_connector = None               # type: Optional[VirtualSegmentDisplayConnector]
        self._text_stack = {}                       # type: Dict[str, TextStackEntry]
        self._current_placeholder = None            # type: Optional[TextTemplate]
        self._current_placeholder_future = None     # type: Optional[Future]
        self._current_text_stack_entry = None       # type: Optional[TextStackEntry]
        self._transition_update_task = None         # type: Optional[PeriodicTask]
        self._current_transition = None             # type: Optional[TransitionRunner]
        self._default_color = None                  # type: Optional[RGBColor]
        self._current_state = None                  # type: Optional[SegmentDisplayState]

    async def _initialize(self):
        """Initialize display."""
        await super()._initialize()
        # load platform
        self.platform = self.machine.get_platform_sections('segment_displays', self.config['platform'])
        self.platform.assert_has_feature("segment_displays")

        if not self.platform.features['allow_empty_numbers'] and self.config['number'] is None:
            self.raise_config_error("Segment Display must have a number.", 1)

        self.size = self.config['size']
        self._default_color = [RGBColor(color) for color in self.config["default_color"][0:self.size]]
        if len(self._default_color) < self.size:
            self._default_color += [RGBColor("white")] * (self.size - len(self._default_color))

        # configure hardware
        try:
            self.hw_display = await self.platform.configure_segment_display(self.config['number'],
                                                                            self.size,
                                                                            self.config['platform_settings'])
        except AssertionError as ex:
            raise AssertionError("Error in platform while configuring segment display {}. "
                                 "See error above.".format(self.name)) from ex

        text = SegmentDisplayText.from_str("", self.size, self.config['integrated_dots'],
                                           self.config['integrated_commas'], self.config['use_dots_for_commas'],
                                           self._default_color)
        self._update_display(SegmentDisplayState(text, FlashingType.NO_FLASH, ''))

    def add_virtual_connector(self, virtual_connector):
        """Add a virtual connector instance to connect this segment display to the MPF-MC for virtual displays."""
        self.virtual_connector = virtual_connector

    def remove_virtual_connector(self):
        """Remove the virtual connector instance from this segment display."""
        self.virtual_connector = None

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None) -> dict:
        """Return the parsed and validated config.

        Args:
        ----
            config: Config of device
            is_mode_config: Whether this device is loaded in a mode or system-wide
            debug_prefix: Prefix to use when logging.

        Returns: Validated config
        """
        config = super().validate_and_parse_config(config, is_mode_config, debug_prefix)
        platform = self.machine.get_platform_sections('segment_displays', config.get("platform", None))
        platform.assert_has_feature("segment_displays")
        config['platform_settings'] = platform.validate_segment_display_section(self,
                                                                                config.get('platform_settings', None))

        # For backwards compatibility, an undefined update_method config will be mapped
        # to "stack". However the default in a future version will be "replace", so inform
        # the user that they should explicitly define "stack" to avoid a breaking change.
        if config['update_method'] is None:
            self.warning_log("Segment display update_method will default to 'replace' in a future MPF release. "
                             "You have no config value specified and are currently getting 'stack' for backwards "
                             "compatibility. To avoid a breaking change, please set 'update_method: \"stack\"' "
                             "in your segment display config.")
            config['update_method'] = "stack"
        return config

    # pylint: disable-msg=too-many-arguments
    def add_text_entry(self, text, color, flashing, flash_mask, transition, transition_out, priority, key):
        """Add text to display stack, or send it directly to the display.

        This will replace texts with the same key.
        """
        if self.config['update_method'] == "stack":
            self._text_stack[key] = TextStackEntry(
                text, color, flashing, flash_mask, transition, transition_out, priority, key)
            self._update_stack()
            return

        if self.config['update_method'] != "replace":
            raise ValueError(f"Unknown update_method '{self.config['update_method']}' for segment display {self.name}")

        # For the replace-text update method, skip the stack and write straight to the display
        new_text = TextTemplate(self.machine, text).evaluate({})
        text = SegmentDisplayText.from_str(new_text, self.size, self.config['integrated_dots'],
                                           self.config['integrated_commas'], self.config['use_dots_for_commas'],
                                           color)
        self._update_display(SegmentDisplayState(text, flashing, flash_mask))

    def add_text(self, text: str, priority: int = 0, key: str = None) -> None:
        """Add text to display stack.

        This will replace texts with the same key.
        """
        self.add_text_entry(text, None, None, None, None, None, priority, key)

    def remove_text_by_key(self, key: Optional[str]):
        """Remove entry from text stack."""
        if self.config['update_method'] != "stack":
            self.info_log("Segment display 'remove' action is TBD.")
            return

        if key in self._text_stack:
            del self._text_stack[key]
            self._update_stack()

    # pylint: disable=too-many-arguments
    def _start_transition(self, transition: TransitionBase, current_text: str, new_text: str,
                          current_colors: List[RGBColor], new_colors: List[RGBColor],
                          update_hz: float, flashing, flash_mask):
        """Start the specified transition."""
        current_colors = self._expand_colors(current_colors, len(current_text))
        new_colors = self._expand_colors(new_colors, len(new_text))
        if self._current_transition:
            self._stop_transition()
        self._current_transition = TransitionRunner(self.machine, transition, current_text, new_text,
                                                    current_colors, new_colors)
        transition_text = next(self._current_transition)
        self._update_display(SegmentDisplayState(transition_text, flashing, flash_mask))
        self._transition_update_task = self.machine.clock.schedule_interval(self._update_transition, 1 / update_hz)

    def _update_transition(self):
        """Update the current transition (callback function from transition interval clock)."""
        try:
            transition_text = next(self._current_transition)
            self._update_display(SegmentDisplayState(transition_text, self._current_state.flashing,
                                                     self._current_state.flash_mask))

        except StopIteration:
            self._stop_transition()

    def _stop_transition(self):
        """Stop the current transition."""
        if self._transition_update_task:
            self._transition_update_task.cancel()
            self._transition_update_task = None

        if self._current_transition:
            self._current_transition = None

            if self._current_text_stack_entry:
                # update placeholder
                if len(self._current_text_stack_entry.text) > 0:
                    self._current_placeholder = TextTemplate(self.machine, self._current_text_stack_entry.text)
                    self._current_placeholder_changed()
            else:
                self._current_placeholder = None

    def _expand_colors(self, colors, length):
        """Expand color to a certain length."""
        if not colors:
            colors = self._default_color
        if len(colors) > length:
            colors = colors[0:length]
        elif len(colors) < length:
            colors = colors + [colors[len(colors) - 1]] * (length - len(colors))

        return colors

    def _update_stack(self) -> None:
        """Sort stack and show top entry on display."""
        # do nothing if stack is emtpy. set display empty
        assert self.hw_display is not None
        if not self._text_stack:
            top_text_stack_entry = TextStackEntry(" " * self.size, None, FlashingType.NO_FLASH, "", None, None,
                                                  -999999, "")
        else:
            # sort text stack by priority
            self._text_stack = dict(
                sorted(self._text_stack.items(), key=lambda item: item[1].priority, reverse=True))

            # get top entry (highest priority)
            top_text_stack_entry = next(iter(self._text_stack.values()))

        previous_text_stack_entry = self._current_text_stack_entry
        self._current_text_stack_entry = top_text_stack_entry

        # determine if the new key is different than the previous key (out transitions are only applied
        # when changing keys)
        transition_config = None
        if previous_text_stack_entry and top_text_stack_entry.key != previous_text_stack_entry.key and \
                previous_text_stack_entry.transition_out:
            transition_config = previous_text_stack_entry.transition_out

        # determine if new text entry has a transition, if so, apply it (overrides any outgoing transition)
        if top_text_stack_entry.transition:
            transition_config = top_text_stack_entry.transition

        # start transition (if configured)
        if transition_config:
            transition = TransitionManager.get_transition(self.size,
                                                          self.config['integrated_dots'],
                                                          self.config['integrated_commas'],
                                                          self.config['use_dots_for_commas'],
                                                          transition_config)
            if previous_text_stack_entry:
                previous_text = previous_text_stack_entry.text
            else:
                previous_text = " " * self.size

            if top_text_stack_entry.flashing is not None:
                flashing = top_text_stack_entry.flashing
                flash_mask = top_text_stack_entry.flash_mask
            else:
                flashing = self._current_state.flashing
                flash_mask = self._current_state.flash_mask

            self._start_transition(transition, previous_text, top_text_stack_entry.text,
                                   self._current_state.text.get_colors(), top_text_stack_entry.colors,
                                   self.config['default_transition_update_hz'], flashing, flash_mask)
        else:
            # no transition - subscribe to text template changes and update display
            self._current_placeholder = TextTemplate(self.machine, top_text_stack_entry.text)
            new_text, future = self._current_placeholder.evaluate_and_subscribe({})
            future.add_done_callback(self._current_placeholder_changed)

            # set any flashing state specified in the entry
            if top_text_stack_entry.flashing is not None:
                flashing = top_text_stack_entry.flashing
                flash_mask = top_text_stack_entry.flash_mask
            else:
                flashing = self._current_state.flashing
                flash_mask = self._current_state.flash_mask

            # update colors if specified
            if top_text_stack_entry.colors:
                colors = top_text_stack_entry.colors
            else:
                colors = self._current_state.text.get_colors()

            # update the display
            text = SegmentDisplayText.from_str(new_text, self.size, self.config['integrated_dots'],
                                               self.config['integrated_commas'], self.config['use_dots_for_commas'],
                                               colors)
            self._update_display(SegmentDisplayState(text, flashing, flash_mask))

    def _current_placeholder_changed(self, future=None, **kwargs) -> None:
        """Update display when a placeholder changes (callback function)."""
        del kwargs
        if future and future.cancelled():
            return
        if self._current_placeholder_future and not self._current_placeholder_future.done():
            self._current_placeholder_future.cancel()
        new_text, self._current_placeholder_future = self._current_placeholder.evaluate_and_subscribe({})
        self._current_placeholder_future.add_done_callback(self._current_placeholder_changed)
        text = SegmentDisplayText.from_str(new_text, self.size, self.config['integrated_dots'],
                                           self.config['integrated_commas'], self.config['use_dots_for_commas'],
                                           self._current_state.text.get_colors())
        self._update_display(SegmentDisplayState(text, self._current_state.flashing,
                                                 self._current_state.flash_mask))

    def set_flashing(self, flashing: FlashingType, flash_mask: str = ""):
        """Enable/Disable flashing."""
        self._update_display(SegmentDisplayState(self._current_state.text, flashing, flash_mask))

    def set_color(self, colors: List[RGBColor]):
        """Set display colors."""
        assert isinstance(colors, list)
        assert self.hw_display is not None
        text = SegmentDisplayText.from_str(self._current_state.text.convert_to_str(), self.size,
                                           self.config['integrated_dots'], self.config['integrated_commas'],
                                           self.config['use_dots_for_commas'], colors)
        self._update_display(SegmentDisplayState(text,
                                                 self._current_state.flashing,
                                                 self._current_state.flash_mask))

    @property
    def text(self):
        """Return current text."""
        return self._current_state.text.convert_to_str()

    @property
    def colors(self):
        """Return current colors."""
        return self._current_state.text.get_colors()

    @property
    def flashing(self):
        """Return current flashing state."""
        return self._current_state.flashing

    @property
    def flash_mask(self):
        """Return current flash mask."""
        return self._current_state.flash_mask

    def _update_display(self, new_state: SegmentDisplayState) -> None:
        """Update display to current text."""
        assert self.hw_display is not None
        if self._current_state and new_state == self._current_state:
            return

        self._current_state = new_state
        text = new_state.text
        flashing = new_state.flashing
        flash_mask = new_state.flash_mask

        assert len(text) == self.size

        # set text to display
        self.hw_display.set_text(text, flashing=flashing, flash_mask=flash_mask)
        if self.virtual_connector:
            self.virtual_connector.set_text(self.name, text, flashing=flashing, flash_mask=flash_mask)



"""OPP servo implementation."""

from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface
from mpf.platforms.opp.opp_rs232_intf import OppRs232Intf

MYPY = False
if MYPY:  # pragma: no cover
    from mpf.platforms.opp.opp import OppHardwarePlatform  # pylint: disable-msg=cyclic-import,unused-import


class OPPServo(ServoPlatformInterface):

    """A servo in the OPP platform."""

    __slots__ = ["number", "chain_serial", "platform", "speed", "current_position"]

    def __init__(self, chain_serial, servo_num, platform: "OppHardwarePlatform"):
        """Initialize servo."""
        self.number = servo_num
        self.platform = platform
        self.chain_serial = chain_serial
        self.speed = 0
        self.current_position = 0

    def stop(self):
        """Disable servo.

        Set position to 0 to disable servo.
        """
        self.go_to_position(0)

    def go_to_position(self, position):
        """Set a servo position.

        position [0 to 1.0] is converted to position_numeric [0 to 255].
        position_numeric is measured in 10us intervals, so a position_numeric of 100 is a 1ms pulse
        and 150 is 1.5ms. A position_numeric of 0 disables the servo. Use caution with extreme
        position values as it could cause a servo to drive to a position it cannot reach.
        """
        # convert from [0,1] to [0, 255]
        position_numeric = int(position * 255)
        servo_offset = 0x3000 + self.number

        if position_numeric == 0 or self.current_position == 0 or self.speed <= 0:
            fade_ms = 0
        else:
            fade_ms = 600 * abs(position_numeric - self.current_position) / self.speed
            if fade_ms > 65535:
                fade_ms = 65535

        msg = bytearray()
        msg.append(0x20)
        msg.append(OppRs232Intf.SERIAL_LED_CMD_FADE)
        msg.append(int(servo_offset / 256))
        msg.append(int(servo_offset % 256))
        msg.append(int(0))  # number of servos (high)
        msg.append(int(1))  # number of servos (low)...only commanding one at a time.
        msg.append(int(fade_ms / 256))
        msg.append(int(fade_ms % 256))
        msg.append(position_numeric)
        msg.extend(OppRs232Intf.calc_crc8_whole_msg(msg))
        cmd = bytes(msg)

        self.platform.debug_log("Set servo position on %s: %s", self.chain_serial, "".join(" 0x%02x" % b for b in cmd))

        self.platform.send_to_processor(self.chain_serial, cmd)

        self.current_position = position_numeric

    def set_speed_limit(self, speed_limit):
        """Set the speed of this servo.

        For the standard 1ms pulse width change to move a servo between
        extremes, a speed of 1 will take 1 minute, and a speed of 60 would take
        1 second.

        A speed <= 0 is unrestricted by firmware
        """
        self.platform.debug_log("Change servo speed limit on %s: %s", self.chain_serial, int(speed_limit))

        self.speed = speed_limit

    def set_acceleration_limit(self, acceleration_limit):
        """Not implemented."""



# mpf.tests.test_Fast_Exp

from mpf.core.rgb_color import RGBColor
from mpf.tests.test_Fast import TestFastBase


class TestFastExp(TestFastBase):
    """Tests the FAST EXP boards."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.serial_connections_to_mock = ['net2', 'exp']

    def get_config_file(self):
        return 'exp.yaml'

    def create_expected_commands(self):
        # These are all the defaults based on the config file for this test.
        # Individual tests can override / add as needed

        self.serial_connections['exp'].expected_commands = {'RA@880:000000': '',
                                                            'RA@881:000000': '',
                                                            'RA@882:000000': '',
                                                            'RA@890:000000': '',
                                                            'RA@892:000000': '',
                                                            'RA@B40:000000': '',
                                                            'RA@840:000000': '',
                                                            'RA@841:000000': '',
                                                            'RA@480:000000': '',
                                                            'RA@481:000000': '',
                                                            'RA@482:000000': '',
                                                            'RF@89:5DC': '',
                                                            'EM@B40:0,1,7D0,1F4,9C4,5DC': '',
                                                            'EM@B40:1,1,7D0,3E8,7D0,5DC': '',
                                                            'EM@882:7,1,7D0,3E8,7D0,5DC': '',
                                                            'MP@B40:0,7F,7D0': '',
                                                            'MP@B40:1,7F,7D0': '',
                                                            'MP@882:7,7F,7D0': '',}

    def test_servo(self):
        # go to min position
        self.exp_cpu.expected_commands = {
                "MP@B40:0,00,7D0": ""                    # MP:<INDEX>,<POSITION>,<TIME_MS><CR>
        }
        self.machine.servos["servo1"].go_to_position(0)
        self.advance_time_and_run(1)
        self.assertFalse(self.exp_cpu.expected_commands)

        # go to max position
        self.exp_cpu.expected_commands = {
                "MP@B40:0,FF,7D0": ""
        }
        self.machine.servos["servo1"].go_to_position(1)
        self.advance_time_and_run(.1)
        self.assertFalse(self.exp_cpu.expected_commands)

    def test_leds(self):

        # create local references to all the lights so they can be accessed like `led1.on()`
        for led_name, led_obj in self.machine.lights.items():
            setattr(self, led_name, led_obj)

        self.fast_exp_leds = self.machine.default_platform.fast_exp_leds

        self._test_led_internals()
        self._test_led_colors()
        self._test_exp_board_reset()
        self._test_grb_led()
        self._test_rgbw_leds()
        self._test_led_channels()
        self._test_led_software_fade()
        self._test_lew_hardware_fade()

    def _test_led_internals(self):

        # Make sure the internal LED map is correct
        self.assertIn("88100", self.fast_exp_leds)
        self.assertIn("88001", self.fast_exp_leds)
        self.assertIn("88002", self.fast_exp_leds)
        self.assertIn("88120", self.fast_exp_leds)
        self.assertIn("88121", self.fast_exp_leds)
        self.assertIn("89200", self.fast_exp_leds)

        # Make sure explicit offset declarations work
        self.assertEqual(self.led4.hw_drivers['red'][0].number, '88120-0')
        self.assertEqual(self.led4.hw_drivers['green'][0].number, '88120-1')
        self.assertEqual(self.led4.hw_drivers['blue'][0].number, '88120-2')
        self.assertEqual(self.led5.hw_drivers['red'][0].number, '88121-1')
        self.assertEqual(self.led5.hw_drivers['green'][0].number, '88121-2')
        self.assertEqual(self.led5.hw_drivers['blue'][0].number, '88122-0')
        self.assertEqual(self.led6.hw_drivers['red'][0].number, '89200-2')
        self.assertEqual(self.led6.hw_drivers['green'][0].number, '89201-0')
        self.assertEqual(self.led6.hw_drivers['blue'][0].number, '89201-1')


        # Make sure all the RGBW, channels, previous, and start_channels are working
        self.assertEqual(self.led22.hw_drivers['red'][0].number, '48002-0')
        self.assertEqual(self.led22.hw_drivers['green'][0].number, '48002-1')
        self.assertEqual(self.led22.hw_drivers['blue'][0].number, '48002-2')
        self.assertEqual(self.led23.hw_drivers['red'][0].number, '48003-0')
        self.assertEqual(self.led23.hw_drivers['green'][0].number, '48003-1')
        self.assertEqual(self.led23.hw_drivers['blue'][0].number, '48003-2')
        self.assertEqual(self.led24.hw_drivers['red'][0].number, '48004-0')
        self.assertEqual(self.led24.hw_drivers['green'][0].number, '48004-1')
        self.assertEqual(self.led24.hw_drivers['blue'][0].number, '48004-2')
        self.assertEqual(self.led24.hw_drivers['white'][0].number, '48005-0')
        self.assertEqual(self.led25.hw_drivers['red'][0].number, '48005-1')
        self.assertEqual(self.led25.hw_drivers['green'][0].number, '48005-2')
        self.assertEqual(self.led25.hw_drivers['blue'][0].number, '48006-0')
        self.assertEqual(self.led25.hw_drivers['white'][0].number, '48006-1')
        self.assertEqual(self.led26.hw_drivers['red'][0].number, '48006-2')
        self.assertEqual(self.led26.hw_drivers['green'][0].number, '48007-0')
        self.assertEqual(self.led26.hw_drivers['blue'][0].number, '48007-1')
        self.assertEqual(self.led26.hw_drivers['white'][0].number, '48007-2')
        self.assertEqual(self.led27.hw_drivers['red'][0].number, '48008-0')
        self.assertEqual(self.led27.hw_drivers['green'][0].number, '48008-1')
        self.assertEqual(self.led27.hw_drivers['blue'][0].number, '48008-2')
        self.assertEqual(self.led28.hw_drivers['red'][0].number, '88222-0')
        self.assertEqual(self.led28.hw_drivers['green'][0].number, '88222-1')
        self.assertEqual(self.led28.hw_drivers['blue'][0].number, '88222-2')
        self.assertEqual(self.led28.hw_drivers['white'][0].number, '88223-0')
        self.assertEqual(self.led29.hw_drivers['red'][0].number, '48009-0')
        self.assertEqual(self.led29.hw_drivers['green'][0].number, '48009-2')
        self.assertEqual(self.led29.hw_drivers['blue'][0].number, '48009-1')
        self.assertEqual(self.led29.hw_drivers['white'][0].number, '4800A-2')
        self.assertEqual(self.led30.hw_drivers['red'][0].number, 'B406B-0')
        self.assertEqual(self.led30.hw_drivers['green'][0].number, 'B406B-1')
        self.assertEqual(self.led30.hw_drivers['blue'][0].number, 'B406B-2')
        self.assertEqual(self.led31.hw_drivers['red'][0].number, 'B406C-0')
        self.assertEqual(self.led31.hw_drivers['green'][0].number, 'B406C-1')
        self.assertEqual(self.led31.hw_drivers['blue'][0].number, 'B406C-2')


    def _test_led_colors(self):

        self.exp_cpu.expected_commands = {
            'RD@880:0201ff123402121212': '',
            'RD@881:0100ffffff': '',
            'RD@841:0160ffffff': ','}

        self.led1.on()
        self.led2.color("ff1234")
        self.led3.color("121212")
        self.led19.color("ffffff")  # Verifies that ports 5-8 of the EXP-0081 are using breakout address 1 LED range 00-7F
        self.advance_time_and_run()

        self.assertEqual("FFFFFF", self.exp_cpu.leds['led1'])
        self.assertEqual("121212", self.exp_cpu.leds['led3'])
        self.assertEqual("FFFFFF", self.exp_cpu.leds['led19'])
        self.assertFalse(self.exp_cpu.expected_commands)

        # turn on a LED on a different board that has a hex index too
        self.exp_cpu.expected_commands = {'RD@B40:016affffff': '',}
        self.led18.on()
        self.advance_time_and_run()
        self.assertEqual("FFFFFF", self.exp_cpu.leds['led18'])

        # # test led off
        self.exp_cpu.expected_commands = {'RD@881:0100000000': '',}
        self.led1.off()
        self.advance_time_and_run()
        self.assertEqual("000000", self.exp_cpu.leds['led1'])

        # # test led color
        self.exp_cpu.expected_commands = {'RD@890:010002172a': '',}
        self.led7.color(RGBColor((2, 23, 42)))
        self.advance_time_and_run(1)
        self.assertEqual("02172A", self.exp_cpu.leds['led7'])

    def _test_exp_board_reset(self):
        # verify a board reset turns off the LEDs only on the board addresses

        self.exp_cpu.expected_commands = {
            'RD@881:0100ff1234': '',
            'RD@880:0102467fff': '',
            'RD@B40:016a6a6a6a': '',}

        self.led1.color("ff1234")
        self.led3.color("467fff")
        self.led18.color("6a6a6a")
        self.advance_time_and_run()

        self.exp_cpu.write(b'BR@B40:')
        self.advance_time_and_run()

        self.assertEqual("000000", self.exp_cpu.leds['led18'])  # this is on the active board and should be off
        self.assertEqual("FF1234", self.exp_cpu.leds['led1'])  # this is on a non-active board ans should still be on
        self.assertEqual("467FFF", self.exp_cpu.leds['led3'])

    def _test_grb_led(self):
        # test led10 grb
        self.exp_cpu.expected_commands = {'RD@B40:014212ff34': '',}
        self.led10.color("ff1234")
        self.advance_time_and_run()
        self.assertEqual("12FF34", self.exp_cpu.leds['led10'])  # ensure the hardware received the colors in RGB order

    def _test_rgbw_leds(self):
        self.exp_cpu.expected_commands = {'RD@480:02050000000600ff00': '',}
        self.led25.color("ffffff")
        self.advance_time_and_run()

        self.exp_cpu.expected_commands = {'RD@882:022200112223110000': '',}
        self.led28.color("112233")
        self.advance_time_and_run()

    def _test_led_channels(self):
        # LED 29 has random ordered channels
        # red 48009-0, green 48009-2, blue 48009-1, white 48010-2

        # white = 00 00 00 FF which becomes
        # 09 [00 00 00] 0A [00 00 FF]
        self.exp_cpu.expected_commands = {'RD@480:02090000000a0000ff': '',}
        self.led29.color("ffffff")
        self.advance_time_and_run()

        self.exp_cpu.expected_commands = {'RD@480:02090022110a000022': '',}
        self.led29.color("223344")  # -> 00112222
        self.advance_time_and_run()

    def _test_led_software_fade(self):

        self.exp_cpu.expected_commands = {'RD@B40:0169151515': '',
                                          'RD@B40:01692b2b2b': '',
                                          'RD@B40:0169424242': '',
                                          'RD@B40:0169585858': '',
                                          'RD@B40:0169646464': '',}

        self.led17.color(RGBColor((100, 100, 100)), fade_ms=150)
        self.advance_time_and_run(.04)
        self.assertTrue(10 < int(self.exp_cpu.leds['led17'][0:2], 16) < 40)
        self.advance_time_and_run(.04)
        self.assertTrue(30 < int(self.exp_cpu.leds['led17'][0:2], 16) < 60)
        self.advance_time_and_run(.04)
        self.assertTrue(60 < int(self.exp_cpu.leds['led17'][0:2], 16) < 90)
        self.advance_time_and_run(2)
        self.assertEqual("646464", self.exp_cpu.leds['led17'])

    def _test_lew_hardware_fade(self):
        # This is also tested via the config file and the expected commands
        self.exp_cpu.expected_commands = {'RF@88:3E8': '',}
        self.machine.default_platform.exp_boards_by_name["brian"].set_led_fade(1000)
        self.advance_time_and_run()


import unittest
from ruamel.yaml.constructor import DuplicateKeyError
from mpf.file_interfaces.yaml_interface import YamlInterface


class TestYamlInterface(unittest.TestCase):

    def test_duplicate_key(self):
        yaml_str = '''
a: 1
b: 2
a: 3
'''
        with self.assertRaises(DuplicateKeyError):
            YamlInterface.process(yaml_str)

    def test_yaml_interface(self):

        config = """

str_1: "+1"
str_2: "032"
str_3: on
str_4: off
str_5: "123e45"
str_6: hi
str_7: 2:10
str_8: 2:10.1
bool_3: true
bool_4: false
bool_5: True
bool_6: False
int_1: 123

        """

        values = {
            "str_1": "+1",
            "str_2": "032",
            "str_3": "on",
            "str_4": "off",
            "str_5": "123e45",
            "str_6": "hi",
            "str_7": "2:10",
            "str_8": "2:10.1",
            "bool_3": True,
            "bool_4": False,
            "bool_5": True,
            "bool_6": False,
            "int_1": 123,
        }

        parsed_config = YamlInterface.process(config)

        for k, v in parsed_config.items():
            if not type(v) is eval(k.split('_')[0]):
                raise AssertionError('YAML value "{}" is {}, not {}'.format(v,
                    type(v), eval(k.split('_')[0])))
            self.assertEqual(values[k], v)



"""Segment displays on light drivers."""
import logging

from mpf.devices.segment_display.segment_display_text import ColoredSegmentDisplayText
from mpf.core.segment_mappings import SEVEN_SEGMENTS, BCD_SEGMENTS, FOURTEEN_SEGMENTS, SIXTEEN_SEGMENTS,\
    EIGHT_SEGMENTS, TextToSegmentMapper
from mpf.platforms.interfaces.segment_display_platform_interface import SegmentDisplaySoftwareFlashPlatformInterface
from mpf.core.platform import SegmentDisplaySoftwareFlashPlatform


class LightSegmentDisplay(SegmentDisplaySoftwareFlashPlatformInterface):

    """Segment display which drives lights."""

    __slots__ = ["_lights", "_key", "_segment_map", "_current_text"]

    def __init__(self, number, lights, segment_type):
        """Initialize segment display."""
        super().__init__(number)
        self._lights = lights
        if segment_type == "7segment":
            self._segment_map = SEVEN_SEGMENTS
        elif segment_type == "bcd":
            self._segment_map = BCD_SEGMENTS
        elif segment_type == "8segment":
            self._segment_map = EIGHT_SEGMENTS
        elif segment_type == "14segment":
            self._segment_map = FOURTEEN_SEGMENTS
        elif segment_type == "16segment":
            self._segment_map = SIXTEEN_SEGMENTS
        else:
            raise AssertionError("Invalid segment type {}".format(segment_type))

        self._key = "segment_display_{}".format(number)
        self._current_text = None

    def _set_text(self, text: ColoredSegmentDisplayText) -> None:
        """Set text to lights."""
        # get the last chars for the number of chars we have
        assert not text.embed_commas
        text = text[-len(self._lights):]
        if text != self._current_text:
            self._current_text = text
            self._update_text()

    def _update_text(self):
        # iterate lights and chars
        mapped_text = TextToSegmentMapper.map_segment_text_to_segments_with_color(
            self._current_text, len(self._lights), self._segment_map)

        for char, lights_for_char in zip(mapped_text, self._lights):
            for name, light in lights_for_char.items():
                if getattr(char[0], name):
                    light.color(color=char[1], key=self._key)
                else:
                    light.remove_from_stack_by_key(key=self._key)


class LightSegmentDisplaysPlatform(SegmentDisplaySoftwareFlashPlatform):

    """Platform which drives segment displays on lights of another platform."""

    __slots__ = ["config"]

    def __init__(self, machine):
        """Initialize platform."""
        super().__init__(machine)
        self.log = logging.getLogger('Light Segment Displays')
        self.log.debug("Configuring Light Segment Displays")
        self.config = self.machine.config_validator.validate_config("light_segment_displays",
                                                                    self.machine.config.get("light_segment_displays"))

    @classmethod
    def get_segment_display_config_section(cls):
        """Return addition config section for segment displays."""
        return "light_segment_displays_device"

    async def configure_segment_display(self, number: str, display_size: int, platform_settings) -> LightSegmentDisplay:
        """Configure light segment display."""
        del display_size

        if platform_settings['lights'] != []:
            _lights = platform_settings['lights']
        else:
            #currently supporting 14segment displays
            segments = ['a', 'b', 'c', 'd', 'e', 'f', 'g1', 'g2', 'h', 'j', 'k', 'l', 'm', 'n', 'dp']
            digit_len = len(segments)

            #get single list of all lights
            _lights = []
            for lt_group in platform_settings['light_groups']:
                await lt_group.wait_for_loaded()
                _lights.append(lt_group.lights)
                self.log.debug("last light in group %s", lt_group.lights[-1].name)
            _lights = [num for sublist in _lights for num in sublist]

            #split list into dicts of digits
            _lights = [_lights[i:i + digit_len] for i in range(0, len(_lights), digit_len)]
            _lights = [dict(zip(segments, digit_lights)) for digit_lights in _lights]

        display = LightSegmentDisplay(number,
                                      lights=_lights,
                                      segment_type=platform_settings['type'])
        self._handle_software_flash(display)
        return display



"""Test show pools."""
from mpf.tests.MpfTestCase import MpfTestCase, patch


class TestShowPools(MpfTestCase):

    def get_config_file(self):
        return 'test_show_pools.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/shows/'

    def test_pool_random(self):
        with patch("mpf.core.assets.random.randint") as rand:
            rand.return_value = 1
            self.post_event("play_pool_random")
            rand.assert_called_with(1, 4)
        self.assertIn("pool_random", self.machine.show_player.instances["_global"]["show_player"])

        self.assertEqual("leds_name_token",
                         self.machine.show_player.instances["_global"]["show_player"]["pool_random"].name)

        self.post_event("stop_pool_random")
        self.assertNotIn("pool_random", self.machine.show_player.instances["_global"]["show_player"])

        with patch("mpf.core.assets.random.randint") as rand:
            rand.return_value = 2
            self.post_event("play_pool_random")
            rand.assert_called_with(1, 4)
        self.assertIn("pool_random", self.machine.show_player.instances["_global"]["show_player"])

        self.assertEqual("leds_single_color",
                         self.machine.show_player.instances["_global"]["show_player"]["pool_random"].name)

        self.post_event("stop_pool_random")
        self.assertNotIn("pool_random", self.machine.show_player.instances["_global"]["show_player"])

        with patch("mpf.core.assets.random.randint") as rand:
            rand.return_value = 3
            self.post_event("play_pool_random")
            rand.assert_called_with(1, 4)
        self.assertIn("pool_random", self.machine.show_player.instances["_global"]["show_player"])

        self.assertEqual("leds_color_token",
                         self.machine.show_player.instances["_global"]["show_player"]["pool_random"].name)

        self.post_event("stop_pool_random")
        self.assertNotIn("pool_random", self.machine.show_player.instances["_global"]["show_player"])

        with patch("mpf.core.assets.random.randint") as rand:
            rand.return_value = 4
            self.post_event("play_pool_random")
            rand.assert_called_with(1, 4)
        self.assertIn("pool_random", self.machine.show_player.instances["_global"]["show_player"])

        self.assertEqual("leds_extended",
                         self.machine.show_player.instances["_global"]["show_player"]["pool_random"].name)



# mpf.tests.test_Fast_Neuron

from mpf.core.platform import SwitchConfig
from mpf.tests.test_Fast import TestFastBase
from mpf.tests.MpfTestCase import MagicMock


class TestFastNeuron(TestFastBase):
    """Tests FAST Neuron hardware."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.serial_connections_to_mock = ['net2', 'seg', 'dmd']

    def get_config_file(self):
        return 'neuron.yaml'

    def create_expected_commands(self):
        # These is everything that happens based on this config file before the
        # code in the test starts. (These commands also do lots of tests themselves.),
        # including initial switch and driver states.

        super().create_expected_commands()

        net_commands_from_this_config = {
            'NN:00': 'NN:00,FP-I/O-3208-3   ,01.10,08,20,00,00,00,00,00,00',
            'NN:01': 'NN:01,FP-I/O-0804-3   ,01.10,04,08,00,00,00,00,00,00',
            'NN:02': 'NN:02,FP-I/O-1616-3   ,01.10,10,10,00,00,00,00,00,00',
            'NN:03': 'NN:03,FP-I/O-0024-3   ,01.10,08,18,00,00,00,00,00,00',

            # All 104 switches are initialized, even if they do not exist in the MPF config
            "SL:00,01,04,04": "SL:P",
            "SL:01,01,04,04": "SL:P",
            "SL:02,01,04,04": "SL:P",
            "SL:03,02,04,04": "SL:P",
            "SL:04,01,04,04": "SL:P",
            "SL:05,02,04,04": "SL:P",
            "SL:06,01,04,04": "SL:P",
            "SL:07,01,02,02": "SL:P",
            "SL:08,01,04,04": "SL:P",
            "SL:09,01,05,1A": "SL:P",
            "SL:0A,01,04,04": "SL:P",
            "SL:0B,02,02,04": "SL:P",
            "SL:0C,00,00,00": "SL:P",
            "SL:0D,00,00,00": "SL:P",
            "SL:0E,00,00,00": "SL:P",
            "SL:0F,00,00,00": "SL:P",
            "SL:10,00,00,00": "SL:P",
            "SL:11,00,00,00": "SL:P",
            "SL:12,00,00,00": "SL:P",
            "SL:13,00,00,00": "SL:P",
            "SL:14,00,00,00": "SL:P",
            "SL:15,00,00,00": "SL:P",
            "SL:16,00,00,00": "SL:P",
            "SL:17,00,00,00": "SL:P",
            "SL:18,00,00,00": "SL:P",
            "SL:19,00,00,00": "SL:P",
            "SL:1A,00,00,00": "SL:P",
            "SL:1B,00,00,00": "SL:P",
            "SL:1C,00,00,00": "SL:P",
            "SL:1D,00,00,00": "SL:P",
            "SL:1E,00,00,00": "SL:P",
            "SL:1F,00,00,00": "SL:P",
            "SL:20,00,00,00": "SL:P",
            "SL:21,00,00,00": "SL:P",
            "SL:22,00,00,00": "SL:P",
            "SL:23,00,00,00": "SL:P",
            "SL:24,00,00,00": "SL:P",
            "SL:25,00,00,00": "SL:P",
            "SL:26,00,00,00": "SL:P",
            "SL:27,00,00,00": "SL:P",
            "SL:28,01,04,04": "SL:P",
            "SL:29,00,00,00": "SL:P",
            "SL:2A,00,00,00": "SL:P",
            "SL:2B,00,00,00": "SL:P",
            "SL:2C,00,00,00": "SL:P",
            "SL:2D,00,00,00": "SL:P",
            "SL:2E,00,00,00": "SL:P",
            "SL:2F,00,00,00": "SL:P",
            "SL:30,00,00,00": "SL:P",
            "SL:31,00,00,00": "SL:P",
            "SL:32,00,00,00": "SL:P",
            "SL:33,00,00,00": "SL:P",
            "SL:34,00,00,00": "SL:P",
            "SL:35,00,00,00": "SL:P",
            "SL:36,00,00,00": "SL:P",
            "SL:37,00,00,00": "SL:P",
            "SL:38,01,04,04": "SL:P",
            "SL:39,00,00,00": "SL:P",
            "SL:3A,00,00,00": "SL:P",
            "SL:3B,00,00,00": "SL:P",
            "SL:3C,00,00,00": "SL:P",
            "SL:3D,00,00,00": "SL:P",
            "SL:3E,00,00,00": "SL:P",
            "SL:3F,00,00,00": "SL:P",
            "SL:40,00,00,00": "SL:P",
            "SL:41,00,00,00": "SL:P",
            "SL:42,00,00,00": "SL:P",
            "SL:43,00,00,00": "SL:P",
            "SL:44,00,00,00": "SL:P",
            "SL:45,00,00,00": "SL:P",
            "SL:46,00,00,00": "SL:P",
            "SL:47,00,00,00": "SL:P",
            "SL:48,00,00,00": "SL:P",
            "SL:49,00,00,00": "SL:P",
            "SL:4A,00,00,00": "SL:P",
            "SL:4B,00,00,00": "SL:P",
            "SL:4C,00,00,00": "SL:P",
            "SL:4D,00,00,00": "SL:P",
            "SL:4E,00,00,00": "SL:P",
            "SL:4F,00,00,00": "SL:P",
            "SL:50,00,00,00": "SL:P",
            "SL:51,00,00,00": "SL:P",
            "SL:52,00,00,00": "SL:P",
            "SL:53,00,00,00": "SL:P",
            "SL:54,00,00,00": "SL:P",
            "SL:55,00,00,00": "SL:P",
            "SL:56,00,00,00": "SL:P",
            "SL:57,00,00,00": "SL:P",
            "SL:58,00,00,00": "SL:P",
            "SL:59,00,00,00": "SL:P",
            "SL:5A,00,00,00": "SL:P",
            "SL:5B,00,00,00": "SL:P",
            "SL:5C,00,00,00": "SL:P",
            "SL:5D,00,00,00": "SL:P",
            "SL:5E,00,00,00": "SL:P",
            "SL:5F,00,00,00": "SL:P",
            "SL:60,00,00,00": "SL:P",
            "SL:61,00,00,00": "SL:P",
            "SL:62,00,00,00": "SL:P",
            "SL:63,00,00,00": "SL:P",
            "SL:64,00,00,00": "SL:P",
            "SL:65,00,00,00": "SL:P",
            "SL:66,00,00,00": "SL:P",
            "SL:67,00,00,00": "SL:P",

            # Drivers from the config will be initialized with their specific settings
            "DL:00,81,00,10,0A,FF,00,00,00": "DL:P",
            "DL:01,81,00,10,0A,FF,00,FF,00": "DL:P",
            "DL:02,81,00,10,17,AA,00,00,00": "DL:P",
            "DL:05,81,00,10,0A,FF,00,00,1B": "DL:P",
            "DL:06,81,00,70,0A,FF,14,EE,00": "DL:P",
            "DL:07,81,00,10,0A,FF,00,88,00": "DL:P",
            "DL:08,81,00,70,0A,FF,C8,EE,00": "DL:P",
            "DL:0A,81,00,10,18,FE,14,AA,00": "DL:P",
            "DL:0B,81,00,10,14,AA,14,AA,00": "DL:P",
            "DL:0D,81,00,10,0A,FF,00,00,00": "DL:P",
            "DL:0E,81,00,10,0A,FF,00,FF,00": "DL:P",
            "DL:0F,81,00,10,0E,FF,00,01,00": "DL:P",
            "DL:10,81,00,10,0A,FF,00,00,00": "DL:P",
            "DL:11,81,00,10,0A,FF,00,00,00": "DL:P",
            "DL:12,81,00,10,0F,FF,00,00,00": "DL:P",
            "DL:13,81,00,10,0A,FF,00,FF,00": "DL:P",
            "DL:14,81,00,10,11,FF,00,00,00": "DL:P",
            "DL:15,81,00,10,0A,FF,00,FF,00": "DL:P",
            }

        self.net_cpu.expected_commands.update(net_commands_from_this_config)

    def test_coils(self):
        # The default expected commands will verify all the coils are configured properly.
        # We just need to ensure things get enabled properly.
        self.confirm_commands()

        self._test_pulse()
        self._test_long_pulse()
        self._test_timed_enable()
        self._test_default_timed_enable()
        self._test_enable_exception()
        self._test_allow_enable()

        # test hardware scan
        info_str = (
            'NET: FP-CPU-2000 v02.13\n'
            '\n'
            'I/O Boards:\n'
            'Board 0 - Model: FP-I/O-3208, Firmware: 01.10, Switches: 32, Drivers: 8\n'
            'Board 1 - Model: FP-I/O-0804, Firmware: 01.10, Switches: 8, Drivers: 4\n'
            'Board 2 - Model: FP-I/O-1616, Firmware: 01.10, Switches: 16, Drivers: 16\n'
            'Board 3 - Model: FP-I/O-0024, Firmware: 01.10, Switches: 24, Drivers: 8\n'
            )

        self.assertEqual(info_str, self.machine.default_platform.get_info_string())

    def _test_pulse(self):

        coil = self.machine.coils["c_baseline"]

        # pulse based on its initial config
        self.net_cpu.expected_commands = {"TL:00,01": "TL:P"}
        coil.pulse()
        self.confirm_commands()

        # pulse with a non-standard pulse_ms, trigger 89 also pulses now
        self.net_cpu.expected_commands = {'DL:00,89,00,10,32,FF,00,00,00': 'DL:P'}
        coil.pulse(50)
        self.confirm_commands()

        # Pulse again and it should just use a TL since the coil is already configured
        self.net_cpu.expected_commands = {"TL:00,01": "TL:P"}
        coil.pulse(50)
        self.confirm_commands()

        # Pulse default and it should reconfigure to default
        self.net_cpu.expected_commands = {'DL:00,89,00,10,0A,FF,00,00,00': 'DL:P'}
        coil.pulse()
        self.confirm_commands()

        # pulse with non-standard ms and power
        self.net_cpu.expected_commands = {'DL:00,89,00,10,64,92,00,00,00': 'DL:P'}
        coil.pulse(100, 0.375)
        self.confirm_commands()

        # Do that same pulse again and it should just use a TL since the coil is already configured
        self.net_cpu.expected_commands = {"TL:00,01": "TL:P"}
        coil.pulse(100, 0.375)
        self.confirm_commands()

    def _test_long_pulse(self):

        coil = self.machine.coils["c_long_pwm2"]

        # pulse based on its initial config
        self.net_cpu.expected_commands = {"TL:06,01": "TL:P"}
        coil.pulse()
        self.confirm_commands()

        self.advance_time_and_run(21)
        # pulse it again, but disable it partway through

        self.net_cpu.expected_commands = {"TL:06,01": "TL:P",
                                                             "TL:06,02": "TL:P",
                                                            }

        coil.pulse()
        self.advance_time_and_run(1)
        coil.disable()
        self.confirm_commands()

    def _test_timed_enable(self):

        coil = self.machine.coils["c_long_pwm2"]  # DL:06,81,00,70,0A,FF,14,EE,00

        # timed_enable based on its current config
        self.net_cpu.expected_commands = {"TL:06,01": "TL:P"}
        coil.timed_enable()
        self.confirm_commands()

        self.net_cpu.expected_commands = {"DL:06,89,00,70,0F,FF,0A,88,00": "DL:P"}
        coil.timed_enable(1000, 0.25, 15, 1.0)
        self.confirm_commands()

    def _test_default_timed_enable(self):
        # test that a regular pulse() command will use the long pulse config
        coil = self.machine.coils["c_longer_pwm2"]  # DL:08,81,00,70,0A,FF,C8,EE,00

        # timed_enable based on its current config
        self.net_cpu.expected_commands = {"TL:08,01": "TL:P"}
        coil.pulse()
        self.confirm_commands()

    def _test_enable_exception(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.coils["c_baseline"].enable()
            self.advance_time_and_run(.1)

    def _test_allow_enable(self):
        self.net_cpu.expected_commands = {
            "DL:01,C1,00,18,0A,FF,FF,00,00": "DL:P"
        }
        self.machine.coils["c_allow_enable"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def test_autofire_rules(self):
        self._test_pulse_rules()
        self._test_pulse_rules_inverted_switch()
        self._test_long_pulse_rules()
        self._test_enable_exception_hw_rule()
        self._test_manual_action_during_active_rule()

    def _test_pulse_rules(self):
        # Basic pulse rule
        # DL:10,81,00,10,0A,FF,00,00,00
        # SL:28,01,04,04

        # Enable, should just update existing rule via TL but add the switch ID
        self.net_cpu.expected_commands = {"TL:10,00,28": "TL:P"}
        self.machine.autofire_coils["ac_baseline"].enable()
        self.confirm_commands()

        # Disable, should update existing rule via TL
        self.net_cpu.expected_commands = {"TL:10,02": "TL:P"}
        self.machine.autofire_coils["ac_baseline"].disable()
        self.confirm_commands()

        # Re-enable, should update existing rule via TL
        self.net_cpu.expected_commands = {"TL:10,00": "TL:P"}
        self.machine.autofire_coils["ac_baseline"].enable()
        self.confirm_commands()

    def _test_pulse_rules_inverted_switch(self):
        # Basic pulse rule with inverted switch
        # DL:11,81,00,10,0A,FF,00,00,00
        # SL:05,02,04,04

        # Inverted switch, should send a new rule
        self.net_cpu.expected_commands = {"DL:11,11,05,10,0A,FF,00,00,00": "DL:P"}
        self.machine.autofire_coils["ac_inverted_switch"].enable()
        self.confirm_commands()

        # Disable, should update existing rule via TL
        self.net_cpu.expected_commands = {"TL:11,02": "TL:P"}
        self.machine.autofire_coils["ac_inverted_switch"].disable()
        self.confirm_commands()

        # Re-enable, should update existing rule via TL
        self.net_cpu.expected_commands = {"TL:11,00": "TL:P"}
        self.machine.autofire_coils["ac_inverted_switch"].enable()
        self.confirm_commands()

    def _test_long_pulse_rules(self):
        # Pulse rules with long pulse (Mode 70)
        driver = self.machine.coils["c_long_pwm2"].hw_driver
        switch = self.machine.switches["s_debounce_auto"].hw_switch

        # Verify current configs
        self.assertEqual(driver.get_current_config(), 'DL:06,81,00,70,0A,FF,14,EE,00')
        self.assertEqual(switch.get_current_config(), 'SL:06,01,04,04')

        # Enable, should just update existing rule via TL but add the switch ID
        self.net_cpu.expected_commands = {"TL:06,00,06": "TL:P"}
        self.machine.autofire_coils["ac_2_stage_pwm"].enable()
        self.confirm_commands()

        # Enable again, no new commands should be sent
        self.machine.autofire_coils["ac_2_stage_pwm"].enable()
        self.confirm_commands()

        # Disable, should update existing rule via TL
        self.net_cpu.expected_commands = {"TL:06,02": "TL:P"}
        self.machine.autofire_coils["ac_2_stage_pwm"].disable()
        self.confirm_commands()

        # Disable again, no new commands should be sent
        self.machine.autofire_coils["ac_2_stage_pwm"].disable()
        self.confirm_commands()

        # Re-enable, should update existing rule via TL, want to ensure it keeps mode 70
        self.net_cpu.expected_commands = {"TL:06,00": "TL:P"}
        self.machine.autofire_coils["ac_2_stage_pwm"].enable()
        self.confirm_commands()

        self.assertEqual(driver.get_current_config(), 'DL:06,01,06,70,0A,FF,14,EE,00')

    def _test_enable_exception_hw_rule(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.flippers["f_single_wound"].config['main_coil_overwrite']['hold_power'] = 1.0
            self.machine.flippers["f_single_wound"].enable()

        self.machine.flippers["f_single_wound"].config['main_coil_overwrite']['hold_power'] = None

    def _test_manual_action_during_active_rule(self):
        driver = self.machine.coils["c_pwm2"].hw_driver
        switch = self.machine.switches["s_debounce_quick"].hw_switch

        # Verify current configs
        self.assertEqual(driver.get_current_config(), 'DL:0B,81,00,10,14,AA,14,AA,00')
        self.assertEqual(switch.get_current_config(), 'SL:07,01,02,02')

        # Enable, should just update existing rule via TL but add the switch ID
        self.net_cpu.expected_commands = {"TL:0B,00,07": "TL:P"}
        self.machine.autofire_coils["ac_test_action"].enable()
        self.confirm_commands()

        # Manually pulse the coil, should not affect the rule
        self.net_cpu.expected_commands = {"TL:0B,01": "TL:P"}
        self.machine.coils["c_pwm2"].pulse()
        self.confirm_commands()

        # Disable, should update existing rule via TL
        self.net_cpu.expected_commands = {"TL:0B,02": "TL:P"}
        self.machine.autofire_coils["ac_test_action"].disable()
        self.confirm_commands()

        # Manually pulse the coil, should not affect the rule
        self.net_cpu.expected_commands = {"TL:0B,01": "TL:P"}
        self.machine.coils["c_pwm2"].pulse()
        self.confirm_commands()

        # Re-enable, should update existing rule via TL
        self.net_cpu.expected_commands = {"TL:0B,00": "TL:P"}
        self.machine.autofire_coils["ac_test_action"].enable()
        self.confirm_commands()

        # Disabled the rule
        self.net_cpu.expected_commands = {"TL:0B,02": "TL:P"}
        self.machine.autofire_coils["ac_test_action"].disable()
        self.confirm_commands()

        # Manually pulse the coil with non-standard pulse
        # trigger 89 = bit 0 driver enable + bit 3 one_shot + bit 7 disable switch
        # switch is disabled since the rule is disabled
        self.net_cpu.expected_commands = {"DL:0B,89,07,10,64,AA,14,AA,00": "DL:P"}
        self.machine.coils["c_pwm2"].pulse(100)
        self.confirm_commands()

        # Enable the rule, should send new config since last pulse was non-standard
        self.net_cpu.expected_commands = {"DL:0B,01,07,10,14,AA,14,AA,00": "DL:P"}
        self.machine.autofire_coils["ac_test_action"].enable()
        self.confirm_commands()

        # Send non-standard pulse. Rule is enabled so a second DL command should be sent to reset the rule
        # to the proper config
        # trigger 09 = bit 0 driver enable + bit 3 one_shot, bit 7 is cleared since rule is active
        self.net_cpu.expected_commands = {"DL:0B,09,07,10,FA,AA,14,AA,00": "DL:P"}
        self.machine.coils["c_pwm2"].pulse(250)
        self.confirm_commands()  # This also moves the clock 100ms

        # Jump ahead 251ms and make sure the rule was put back
        self.net_cpu.expected_commands = {"DL:0B,01,07,10,14,AA,14,AA,00": "DL:P"}
        self.advance_time_and_run(0.051) # +51ms
        self.confirm_commands()  # +100ms, 251ms total since manual pulse, so this command should have been sent

        # hold a device on manually, then cancel it, make sure autofire comes back
        # mode 18 (pulse + hold), trigger C1 (bit 7 disable switch, bit 6 manual, bit 1 enable)
        self.net_cpu.expected_commands = {"DL:0B,C1,07,18,14,AA,AA,00,00": "DL:P"}
        self.machine.coils["c_pwm2"].enable()
        self.confirm_commands()

        # Make sure nothing else was sent
        self.advance_time_and_run(2)
        self.confirm_commands()

        # Turn off the hold and make sure the rule comes back
        # Need to ensure we switch from mode 18 back to 10
        self.net_cpu.expected_commands = {"DL:0B,01,07,10,14,AA,14,AA,00": "DL:P"}
        self.machine.coils["c_pwm2"].disable()
        self.confirm_commands

        # Make sure nothing else was sent
        self.advance_time_and_run(2)
        self.confirm_commands()

        # Send a series of manual pulses during an autofire and make sure the rule comes back, but not until
        # the last pulse is done

        # Previous command already re-enabled the autofire
        self.assertEqual(driver.get_current_config(), 'DL:0B,01,07,10,14,AA,14,AA,00')

        # Pulse with manual pulse time
        self.net_cpu.expected_commands = {"DL:0B,09,07,10,C8,AA,14,AA,00": "DL:P"}
        self.machine.coils["c_pwm2"].pulse(200)
        self.confirm_commands()  # +100ms

        # Pulse again before the first pulse is done, TL this time since it's the same pulse as last time
        self.net_cpu.expected_commands = {"TL:0B,01": "TL:P"}
        self.machine.coils["c_pwm2"].pulse(200)
        self.confirm_commands()  # +100ms
        self.net_cpu.expected_commands = {"TL:0B,01": "TL:P"}
        self.machine.coils["c_pwm2"].pulse(200)
        self.confirm_commands()  # +100ms

        # Last pulse will reset the rule after 201ms, so jump past that and verify the rule is back
        self.net_cpu.expected_commands = {"DL:0B,01,07,10,14,AA,14,AA,00": "DL:P"}
        self.advance_time_and_run(0.002)
        self.confirm_commands()

    def _switch_hit_cb(self, **kwargs):
        self.switch_hit = True

    def test_switches(self):
        # Default startup SL commands test / confirm all the variations of the switch configs
        self._test_startup_switches()
        self._test_bad_switch_configs()
        self._test_switch_changes()
        self._test_switch_changes_nc()
        self._test_receiving_sa()

    def _test_startup_switches(self):
        self.assertSwitchState("s_baseline", 1)
        self.assertSwitchState("s_flipper", 0)
        self.assertSwitchState("s_test_nc", 0)  # NC which SA reports active should be inactive

    def _test_bad_switch_configs(self):
        # invalid switch
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_switch('io3208-32', SwitchConfig(name="", debounce='auto', invert=0), {})

        # invalid board
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_switch('brian-0', SwitchConfig(name="", debounce='auto', invert=0), {})

    def _test_switch_changes(self):
        self.assertSwitchState("s_flipper", 0)

        self.switch_hit = False
        self.advance_time_and_run(1)
        self.assertSwitchState("s_flipper_eos", 0)
        self.assertFalse(self.switch_hit)

        self.machine.events.add_handler("s_flipper_eos_active", self._switch_hit_cb)
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"-L:02\r")
        self.advance_time_and_run(1)

        self.assertTrue(self.switch_hit)
        self.assertSwitchState("s_flipper_eos", 1)
        self.switch_hit = False

        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_flipper_eos", 1)

        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"/L:02\r")
        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_flipper_eos", 0)

    def _test_switch_changes_nc(self):
        self.switch_hit = False
        self.advance_time_and_run(1)
        self.assertSwitchState("s_test_nc", 0)
        self.assertFalse(self.switch_hit)

        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"-L:05\r")
        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test_nc", 1)

        self.machine.events.add_handler("s_test_nc_inactive", self._switch_hit_cb)
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"/L:05\r")
        self.advance_time_and_run(1)

        self.assertSwitchState("s_test_nc", 0)
        self.assertTrue(self.switch_hit)
        self.switch_hit = False

    def _test_receiving_sa(self):
        # Receive a random SA command during normal operation with an unexpected switch
        # state and ensure it's processed properly

        # Show s_cab_flipper 0x38 switch on
        self.net_cpu.expected_commands = {
            "SA:" : "SA:0E,2900000000000001000000000000"}

        self.assertSwitchState("s_cab_flipper", 0)

        # Send an SA:
        self.loop.run_until_complete(self.machine.default_platform.get_hw_switch_states(True))
        self.advance_time_and_run()

        self.assertSwitchState("s_cab_flipper", 1)

        # Process a random SA coming in (this shouldn't happen but should work in non async mode)
        # Switch 0x09 is also now active
        self.assertSwitchState("s_debounce_custom", 0)
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"SA:0E,2902000000000001000000000000\r")
        self.advance_time_and_run()
        self.assertSwitchState("s_debounce_custom", 1)

    def test_flipper_single_coil_with_opto_switch(self):
        coil = self.machine.coils["c_flipper_single_wound"]
        hw_driver = coil.hw_driver
        switch = self.machine.switches["s_flipper_opto"].hw_switch
        flipper = self.machine.flippers["f_single_wound"]

        self.assertEqual(hw_driver.get_current_config(), 'DL:0F,81,00,10,0E,FF,00,01,00')
        self.assertEqual(switch.get_current_config(), 'SL:03,02,04,04')

        # flipper rule enable
        # Trigger 11 (bit 0 enable, bit 4 invert switch since it's an opto)
        self.net_cpu.expected_commands = {"DL:0F,11,03,18,0E,FF,01,00,00": "DL:P",}
        flipper.enable()
        self.confirm_commands()

        # manual flip while rule is active
        self.net_cpu.expected_commands = {"TL:0F,01": "TL:P"}
        coil.pulse()
        self.confirm_commands()

        # disable rule (tilt)
        self.net_cpu.expected_commands = {"TL:0F,02": "TL:P"}
        flipper.disable()
        self.confirm_commands()

        # enable again, config has already been sent
        self.net_cpu.expected_commands = {"TL:0F,00": "TL:P"}
        flipper.enable()
        self.confirm_commands()

    def test_flipper_two_coils(self):
        main_coil = self.machine.coils["c_flipper_main"]
        hold_coil = self.machine.coils["c_flipper_hold"]
        main_hw_driver = main_coil.hw_driver
        hold_hw_driver = hold_coil.hw_driver
        switch = self.machine.switches["s_flipper"].hw_switch
        flipper = self.machine.flippers["f_dual_wound"]

        self.assertEqual(main_hw_driver.get_current_config(), 'DL:0D,81,00,10,0A,FF,00,00,00')
        self.assertEqual(hold_hw_driver.get_current_config(), 'DL:0E,81,00,10,0A,FF,00,FF,00')
        self.assertEqual(switch.get_current_config(), 'SL:01,01,04,04')

        # flipper rule enable
        # Trigger 1 (bit 0 enable)
        self.net_cpu.expected_commands = {"DL:0D,01,01,18,0A,FF,00,00,00": "DL:P",
                                                             "DL:0E,01,01,18,0A,FF,FF,00,00": "DL:P",}
        flipper.enable()
        self.confirm_commands()

        # manual flip while rule is active
        self.net_cpu.expected_commands = {"TL:0D,01": "TL:P"}
        main_coil.pulse()
        self.confirm_commands()

        # disable rule (tilt)
        self.net_cpu.expected_commands = {"TL:0D,02": "TL:P",
                                                             "TL:0E,02": "TL:P",}
        flipper.disable()
        self.confirm_commands()

        # enable again, config has already been sent
        self.net_cpu.expected_commands = {"TL:0D,00": "TL:P",
                                                             "TL:0E,00": "TL:P",}
        flipper.enable()
        self.confirm_commands()

    def test_flipper_two_coils_with_eos(self):
        main_coil = self.machine.coils["c_flipper2_main"]
        hold_coil = self.machine.coils["c_flipper2_hold"]
        main_hw_driver = main_coil.hw_driver
        hold_hw_driver = hold_coil.hw_driver
        switch = self.machine.switches["s_cab_flipper"].hw_switch
        eos_switch = self.machine.switches["s_flipper_eos"].hw_switch
        flipper = self.machine.flippers["f_test_hold_eos"]

        self.assertEqual(main_hw_driver.get_current_config(), 'DL:12,81,00,10,0F,FF,00,00,00')
        self.assertEqual(hold_hw_driver.get_current_config(), 'DL:13,81,00,10,0A,FF,00,FF,00')
        self.assertEqual(switch.get_current_config(), 'SL:38,01,04,04')

        # flipper rule enable
        # Trigger 1 (bit 0 enable), Mode 75 (pulse+hold w/cancel), EOS switch 38
        self.net_cpu.expected_commands = {"DL:12,01,38,75,02,0F,00,00,00": "DL:P",
                                                             "DL:13,01,38,18,0A,FF,FF,00,00": "DL:P",}
        flipper.enable()
        self.confirm_commands()

        # manual flip while rule is active, this will send a new DL (since mode 75 isn't a normal pulse mode)
        # followed by another DL after a delay to reconfigure it back to the autofire rule
        self.net_cpu.expected_commands = {"DL:12,09,38,10,0F,FF,00,00,00": "DL:P",
                                                             "DL:12,01,38,75,02,0F,00,00,00": "DL:P",}
        main_coil.pulse()
        self.confirm_commands()

        # disable rule (tilt)
        self.net_cpu.expected_commands = {"TL:12,02": "TL:P",
                                                             "TL:13,02": "TL:P",}
        flipper.disable()
        self.confirm_commands()

        # enable again, config has already been sent
        self.net_cpu.expected_commands = {"TL:12,00": "TL:P",
                                                             "TL:13,00": "TL:P",}
        flipper.enable()
        self.confirm_commands()

    def test_flipper_two_coils_with_nc_eos(self):
        main_coil = self.machine.coils["c_flipper3_main"]
        hold_coil = self.machine.coils["c_flipper3_hold"]
        main_hw_driver = main_coil.hw_driver
        hold_hw_driver = hold_coil.hw_driver
        switch = self.machine.switches["s_flipper3"].hw_switch
        eos_switch = self.machine.switches["s_flipper3_eos_nc"].hw_switch
        flipper = self.machine.flippers["f_eos_nc"]

        self.assertEqual(main_hw_driver.get_current_config(), 'DL:14,81,00,10,11,FF,00,00,00')
        self.assertEqual(hold_hw_driver.get_current_config(), 'DL:15,81,00,10,0A,FF,00,FF,00')
        self.assertEqual(switch.get_current_config(), 'SL:0A,01,04,04')

        # flipper rule enable
        # Trigger 21 (bit 0 enable, bit 4 invert second switch), Mode 75 (pulse+hold w/cancel), EOS switch 38
        self.net_cpu.expected_commands = {"DL:14,21,0A,75,0B,11,00,00,00": "DL:P",
                                           "DL:15,01,0A,18,0A,FF,FF,00,00": "DL:P",}
        flipper.enable()
        self.confirm_commands()

    def test_machine_reset(self):

        # Set the commands that will respond to the query on reset. Some of these are
        # changed from the default so we can simulate the machine in a dirty state

        self.net_cpu.expected_commands = {
            "SL:00": "SL:00,01,04,04",
            "SL:01": "SL:01,01,04,04",
            "SL:02": "SL:02,01,04,04",
            "SL:03": "SL:03,02,04,04",
            "SL:04": "SL:04,01,04,04",
            "SL:05": "SL:05,02,04,04",
            "SL:06": "SL:06,01,04,04",
            "SL:07": "SL:07,01,02,02",
            "SL:08": "SL:08,01,04,04",
            "SL:09": "SL:09,01,05,1A",
            "SL:0A": "SL:0A,01,04,04",
            "SL:0B": "SL:0B,02,02,04",
            "SL:0C": "SL:0C,00,00,00",
            "SL:0D": "SL:0D,00,00,00",
            "SL:0E": "SL:0E,02,00,00",
            "SL:0F": "SL:0F,00,00,00",
            "SL:10": "SL:10,00,02,00",
            "SL:11": "SL:11,00,00,00",
            "SL:12": "SL:12,00,22,00",
            "SL:13": "SL:13,00,00,00",
            "SL:14": "SL:14,00,00,00",
            "SL:15": "SL:15,00,00,00",
            "SL:16": "SL:16,00,00,00",
            "SL:17": "SL:17,00,00,00",
            "SL:18": "SL:18,00,00,00",
            "SL:19": "SL:19,00,00,00",
            "SL:1A": "SL:1A,00,00,00",
            "SL:1B": "SL:1B,00,00,00",
            "SL:1C": "SL:1C,00,00,00",
            "SL:1D": "SL:1D,00,00,00",
            "SL:1E": "SL:1E,00,00,00",
            "SL:1F": "SL:1F,00,00,00",
            "SL:20": "SL:20,00,00,00",
            "SL:21": "SL:21,00,00,00",
            "SL:22": "SL:22,00,00,00",
            "SL:23": "SL:23,00,00,00",
            "SL:24": "SL:24,00,00,00",
            "SL:25": "SL:25,00,00,00",
            "SL:26": "SL:26,00,00,00",
            "SL:27": "SL:27,00,00,00",
            "SL:28": "SL:28,01,04,04",
            "SL:29": "SL:29,00,00,00",
            "SL:2A": "SL:2A,00,00,00",
            "SL:2B": "SL:2B,00,00,00",
            "SL:2C": "SL:2C,00,00,00",
            "SL:2D": "SL:2D,00,00,00",
            "SL:2E": "SL:2E,00,00,00",
            "SL:2F": "SL:2F,00,00,00",
            "SL:30": "SL:30,00,00,00",
            "SL:31": "SL:31,00,00,00",
            "SL:32": "SL:32,00,00,00",
            "SL:33": "SL:33,00,00,00",
            "SL:34": "SL:34,00,00,00",
            "SL:35": "SL:35,00,00,00",
            "SL:36": "SL:36,00,00,00",
            "SL:37": "SL:37,00,00,00",
            "SL:38": "SL:38,01,04,04",
            "SL:39": "SL:39,00,00,00",
            "SL:3A": "SL:3A,00,00,00",
            "SL:3B": "SL:3B,00,00,00",
            "SL:3C": "SL:3C,00,00,00",
            "SL:3D": "SL:3D,00,00,00",
            "SL:3E": "SL:3E,00,00,00",
            "SL:3F": "SL:3F,00,00,00",
            "SL:40": "SL:40,00,00,00",
            "SL:41": "SL:41,00,00,00",
            "SL:42": "SL:42,00,00,00",
            "SL:43": "SL:43,00,00,00",
            "SL:44": "SL:44,00,00,00",
            "SL:45": "SL:45,00,00,00",
            "SL:46": "SL:46,00,00,00",
            "SL:47": "SL:47,00,00,00",
            "SL:48": "SL:48,00,00,00",
            "SL:49": "SL:49,00,00,00",
            "SL:4A": "SL:4A,00,00,00",
            "SL:4B": "SL:4B,00,00,00",
            "SL:4C": "SL:4C,00,00,00",
            "SL:4D": "SL:4D,00,00,00",
            "SL:4E": "SL:4E,00,00,00",
            "SL:4F": "SL:4F,00,00,00",
            "SL:50": "SL:50,00,00,00",
            "SL:51": "SL:51,00,00,00",
            "SL:52": "SL:52,00,00,00",
            "SL:53": "SL:53,00,00,00",
            "SL:54": "SL:54,00,00,00",
            "SL:55": "SL:55,00,00,00",
            "SL:56": "SL:56,00,00,00",
            "SL:57": "SL:57,00,00,00",
            "SL:58": "SL:58,00,00,00",
            "SL:59": "SL:59,00,00,00",
            "SL:5A": "SL:5A,00,00,00",
            "SL:5B": "SL:5B,00,00,00",
            "SL:5C": "SL:5C,00,00,00",
            "SL:5D": "SL:5D,00,00,00",
            "SL:5E": "SL:5E,00,00,00",
            "SL:5F": "SL:5F,00,00,00",
            "SL:60": "SL:60,00,00,00",
            "SL:61": "SL:61,00,00,00",
            "SL:62": "SL:62,00,00,00",
            "SL:63": "SL:63,00,00,00",
            "SL:64": "SL:64,00,00,00",
            "SL:65": "SL:65,00,00,00",
            "SL:66": "SL:66,00,00,00",
            "SL:67": "SL:67,00,00,00",

            # Here's the cleanup from these
            "SL:0E,00,00,00": "SL:P",
            "SL:10,00,00,00": "SL:P",
            "SL:12,00,00,00": "SL:P",

            # Here are the DL queries, some are dirty
            "DL:00": "DL:00,81,00,10,0A,FF,00,00,00",
            "DL:01": "DL:01,81,00,10,0A,FF,00,FF,00",
            "DL:02": "DL:02,81,00,10,17,AA,00,00,00",
            "DL:03": "DL:03,00,00,00,00,00,00,00,00",
            "DL:04": "DL:04,00,00,00,00,00,00,00,00",
            "DL:05": "DL:05,81,00,10,0A,FF,00,00,1B",
            "DL:06": "DL:06,81,00,70,0A,FF,14,EE,00",
            "DL:07": "DL:07,81,00,10,0A,FF,00,88,00",
            "DL:08": "DL:08,81,00,70,0A,FF,C8,EE,00",
            "DL:09": "DL:09,00,00,00,00,00,00,00,00",
            "DL:0A": "DL:0A,81,00,10,18,FE,14,AA,00",
            "DL:0B": "DL:0B,81,00,10,14,AA,14,AA,00",
            "DL:0C": "DL:0C,00,00,00,00,00,00,00,00",
            "DL:0D": "DL:0D,01,01,18,0A,FF,00,00,00",
            "DL:0E": "DL:0E,81,00,10,0A,FF,00,FF,00",
            "DL:0F": "DL:0F,11,03,18,0E,FF,01,00,00",
            "DL:10": "DL:10,81,00,10,0A,FF,00,00,00",
            "DL:11": "DL:11,81,00,10,0A,FF,00,00,00",
            "DL:12": "DL:12,81,00,10,0F,FF,00,00,00",
            "DL:13": "DL:13,81,00,10,0A,FF,00,FF,00",
            "DL:14": "DL:14,81,00,10,11,FF,00,00,00",
            "DL:15": "DL:15,81,00,10,0A,FF,00,FF,00",
            "DL:16": "DL:16,00,00,00,00,00,00,00,00",
            "DL:17": "DL:17,00,00,00,00,00,00,00,00",
            "DL:18": "DL:18,00,00,00,00,00,00,00,00",
            "DL:19": "DL:19,00,00,00,00,00,00,00,00",
            "DL:1A": "DL:1A,00,00,00,00,00,00,00,00",
            "DL:1B": "DL:1B,00,00,00,00,00,00,0A,00",
            "DL:1C": "DL:1C,00,00,00,00,00,00,00,00",
            "DL:1D": "DL:1D,00,00,00,00,00,00,00,00",
            "DL:1E": "DL:1E,00,00,00,00,00,00,00,00",
            "DL:1F": "DL:1F,00,00,00,00,00,00,00,00",
            "DL:20": "DL:20,00,00,00,00,00,00,00,00",
            "DL:21": "DL:21,00,00,00,00,00,00,00,00",
            "DL:22": "DL:22,00,00,00,00,00,00,00,00",
            "DL:23": "DL:23,00,00,00,00,00,00,00,00",
            "DL:24": "DL:24,00,00,00,00,00,00,00,00",
            "DL:25": "DL:25,00,00,00,00,00,00,00,00",
            "DL:26": "DL:26,00,00,00,00,00,00,00,00",
            "DL:27": "DL:27,00,00,00,00,00,00,00,00",
            "DL:28": "DL:28,00,00,00,00,00,00,00,00",
            "DL:29": "DL:29,00,00,00,00,00,00,00,00",
            "DL:2A": "DL:2A,00,00,00,00,00,00,00,00",
            "DL:2B": "DL:2B,00,00,00,00,00,00,00,00",
            "DL:2C": "DL:2C,00,00,00,00,00,00,00,00",
            "DL:2D": "DL:2D,00,00,00,00,00,00,00,00",
            "DL:2E": "DL:2E,00,00,00,00,00,00,00,00",
            "DL:2F": "DL:2F,00,00,00,00,00,00,00,00",

            # Cleaning up the dirty ones
            "DL:0D,81,00,10,0A,FF,00,00,00": "DL:P",
            "DL:0F,81,00,10,0E,FF,00,01,00": "DL:P",
            "DL:1B,00,00,00,00,00,00,00,00": "DL:P",
            }

        # reset the machine and ensure all the dirty devices get reset
        self.loop.run_until_complete(self.machine.reset())
        self.advance_time_and_run()

    def test_fast_game(self):

        self.machine.playfield.add_ball = MagicMock()
        self.machine.ball_controller.num_balls_known = 3

        # Ensure the various autofires get enabled
        self.net_cpu.expected_commands = {
            "TL:10,00,28": "TL:P",                    # ac_baseline
            "DL:11,11,05,10,0A,FF,00,00,00": "DL:P",  # ac_inverted_switch
            "TL:06,00,06": "TL:P",                    # ac_2_stage_pwm
            "TL:0B,00,07": "TL:P",                    # ac_test_action
            "DL:0D,01,01,18,0A,FF,00,00,00": "DL:P",  # f_dual_wound
            "DL:0E,01,01,18,0A,FF,FF,00,00": "DL:P",  # f_dual_wound
            "DL:0F,11,03,18,0E,FF,01,00,00": "DL:P",  # f_single_wound
            "DL:12,01,38,75,02,0F,00,00,00": "DL:P",  # f_test_hold_eos
            "DL:13,01,38,18,0A,FF,FF,00,00": "DL:P",  # f_test_hold_eos
            "DL:14,21,0A,75,0B,11,00,00,00": "DL:P",  # f_eos_nc
            "DL:15,01,0A,18,0A,FF,FF,00,00": "DL:P",  # f_eos_nc
        }

        self.hit_and_release_switch("s_debounce_auto")
        self.advance_time_and_run()
        self.assertIsNotNone(self.machine.game)

        # hit the flipper and to simulate the EOS and everything
        # and earlier bug caused it to disable the flipper which is why we test this now
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"-L:0A\r")  # flipper
        self.advance_time_and_run(0.015)
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"-L:0B\r")  # eos
        self.advance_time_and_run(3)

        # release both
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"/L:0A\r")  # flipper
        self.advance_time_and_run(0.001)
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"/L:0B\r")
        self.advance_time_and_run(3)

        # ---

        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"-L:0A\r")  # flipper
        self.advance_time_and_run(0.016)
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"-L:0B\r")  # eos
        self.advance_time_and_run(0.100)

        # release both
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"/L:0A\r")  # flipper
        self.advance_time_and_run(0.100)
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"/L:0B\r")
        self.advance_time_and_run(3)

        # end game
        self.net_cpu.expected_commands = {
            "TL:10,02": "TL:P",
            "TL:10,02": "TL:P",
            "TL:11,02": "TL:P",
            "TL:06,02": "TL:P",
            "TL:0B,02": "TL:P",
            "TL:0D,02": "TL:P",
            "TL:0E,02": "TL:P",
            "TL:0F,02": "TL:P",
            "TL:12,02": "TL:P",
            "TL:13,02": "TL:P",
            "TL:14,02": "TL:P",
            "TL:15,02": "TL:P",
        }

        self.machine.game.end_game()
        self.advance_time_and_run()

        # Start a game again, all commands should be TLs now

        self.net_cpu.expected_commands = {
            "TL:10,00": "TL:P",
            "TL:11,00": "TL:P",
            "TL:06,00": "TL:P",
            "TL:0B,00": "TL:P",
            "TL:0D,00": "TL:P",
            "TL:0E,00": "TL:P",
            "TL:0F,00": "TL:P",
            "TL:12,00": "TL:P",
            "TL:13,00": "TL:P",
            "TL:14,00": "TL:P",
            "TL:15,00": "TL:P",
        }

        self.hit_and_release_switch("s_debounce_auto")
        self.advance_time_and_run()


# mpf.tests.test_Fast_Audio

from mpf.tests.test_Fast import TestFastBase
from mpf.tests.MpfTestCase import test_config

class TestFastAudio(TestFastBase):
    """Tests the FAST Audio Interface boards."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.serial_connections_to_mock = ['aud']

    def get_config_file(self):
        return 'audio.yaml'

    def create_expected_commands(self):

        if self._testMethodName == 'test_machine_var_loading':
            self.serial_connections['aud'].expected_commands = {
                'AM:0F':'AM:0F',  # all 3 amps enabled, hp is line level
                'AH:11':'AH:11',
                'AV:0F':'AV:0F',
                'AS:0F':'AS:0F',
            }

            self.serial_connections['aud'].autorespond_commands['WD:3E8'] = 'WD:3E8,03'

        else:
            self.serial_connections['aud'].expected_commands = {
                # 'AV:00':'AV:00',
                # 'AS:00':'AS:00',
                # 'AH:00':'AH:00',
                'AM:0B':'AM:0B',
                'AV:08':'AV:08',
                'AS:09':'AS:09',
                'AH:0A':'AH:0A',
            }

    def test_audio_basics(self):

        fast_audio = self.machine.default_platform.audio_interface
        # 31 steps, 0-63
        main_list = [0, 3, 5, 7, 9, 11, 13, 15, 17, 19,
                     22, 24, 26, 28, 30, 32, 34, 36, 38, 40,
                     43, 45, 47, 49, 51, 53, 55, 57, 59, 61, 63]

        # 21 steps, 0-50
        sub_list = [0, 3, 5, 8, 10, 13, 15, 18, 20, 23,
                    25, 28, 30, 33, 35, 38, 40, 43, 45, 48, 50]

        # 11 steps, 0-40
        headphones_list = [0, 4, 8, 12, 16, 20, 24, 28, 32, 36, 40]

        self.assertEqual(main_list, fast_audio.amps['main']['levels_list'])
        self.assertEqual(sub_list, fast_audio.amps['sub']['levels_list'])
        self.assertEqual(headphones_list, fast_audio.amps['headphones']['levels_list'])

        self.assertEqual(8, self.machine.variables.get_machine_var('fast_audio_main_volume'))
        self.assertEqual(9, self.machine.variables.get_machine_var('fast_audio_sub_volume'))
        self.assertEqual(10, self.machine.variables.get_machine_var('fast_audio_headphones_volume'))

        self.assertEqual(8, fast_audio.get_volume('main'))
        self.assertEqual(9, fast_audio.get_volume('sub'))
        self.assertEqual(10, fast_audio.get_volume('headphones'))

        self.assertEqual('08', fast_audio.communicator._volume_to_hw(fast_audio.get_volume('main')))
        self.assertEqual('09', fast_audio.communicator._volume_to_hw(fast_audio.get_volume('sub')))
        self.assertEqual('0A', fast_audio.communicator._volume_to_hw(fast_audio.get_volume('headphones')))

        self.assertEqual('08', self.aud_cpu.main_volume)
        self.assertEqual('09', self.aud_cpu.sub_volume)
        self.assertEqual('0A', self.aud_cpu.headphones_volume)

        self.assertTrue(fast_audio.communicator.amps['main']['enabled'])
        self.assertTrue(fast_audio.communicator.amps['sub']['enabled'])
        self.assertTrue(fast_audio.communicator.amps['headphones']['enabled'])

        # Change the volume var and make sure it's reflected in the hardware
        self.aud_cpu.expected_commands['AV:0D'] = 'AV:0D'
        self.machine.variables.set_machine_var('fast_audio_main_volume', 13)
        self.advance_time_and_run(1)
        self.assertEqual('0D', self.aud_cpu.main_volume)
        self.assertEqual(13, fast_audio.get_volume('main'))
        self.assertEqual('0D', fast_audio.communicator._volume_to_hw(fast_audio.get_volume('main')))

        # Test machine var player events work from config
        self.aud_cpu.expected_commands['AV:0E'] = 'AV:0E'
        self.post_event('increase_main_volume', 1)
        self.assertEqual('0E', self.aud_cpu.main_volume)
        self.assertEqual(14, fast_audio.get_volume('main'))
        self.assertEqual('0E', fast_audio.communicator._volume_to_hw(fast_audio.get_volume('main')))

        # test ducking / restore
        self.assertEqual(14, fast_audio.communicator.amps['main']['volume'])
        self.aud_cpu.expected_commands['AV:13'] = 'AV:13'
        self.post_event('fast_audio_temp_volume', amp_name='main', change=5)
        self.advance_time_and_run(1)
        self.assertEqual(19, fast_audio.communicator.amps['main']['volume'])
        self.assertEqual(14, self.machine.variables.get_machine_var('fast_audio_main_volume'))

        self.aud_cpu.expected_commands['AV:0E'] = 'AV:0E'
        self.post_event('fast_audio_restore', amp_name='main')
        self.advance_time_and_run(1)
        self.assertEqual(14, fast_audio.communicator.amps['main']['volume'])

    def test_control_pins(self):
        self.aud_cpu.expected_commands['XO:01,63'] = 'XO:P'
        self.post_event('fast_audio_pulse_lcd_pin', pin=2)
        self.advance_time_and_run(1)
        self.aud_cpu.expected_commands['XO:06,62'] = 'XO:P'
        self.post_event('fast_audio_pulse_power_pin',1)
        self.aud_cpu.expected_commands['XO:07,61'] = 'XO:P'
        self.post_event('fast_audio_pulse_reset_pin',1)

    @test_config('audio2.yaml')
    def test_machine_var_loading(self):
        fast_audio = self.machine.default_platform.audio_interface
        self.advance_time_and_run()
        self.assertEqual(15, self.machine.variables.get_machine_var('fast_audio_main_volume'))
        self.assertEqual(15, fast_audio.get_volume('main'))
        self.assertEqual(15, fast_audio.communicator.amps['main']['volume'])
        # sub has a machine value set in the configs, which is persisted
        self.assertEqual(2, self.machine.variables.get_machine_var('fast_audio_sub_volume'))
        # sub is linked to main, so it will be 15 even though the config value is 2
        self.assertEqual(15, fast_audio.get_volume('main'))
        self.assertEqual(15, fast_audio.communicator.amps['sub']['volume'])
        self.assertEqual(17, self.machine.variables.get_machine_var('fast_audio_headphones_volume'))



"""Contains the YamlInterface class for reading & writing YAML files."""
import copy
from typing import Any, Iterable, Dict

from ruamel import yaml
from ruamel.yaml.error import MarkedYAMLError

from mpf.core.file_interface import FileInterface

_yaml = yaml.YAML(typ='safe')
_yaml.default_flow_style = False


class YamlInterface(FileInterface):

    """File interface for yaml files."""

    file_types = ['.yaml', '.yml']
    cache = False
    file_cache = dict()     # type: Dict[str, Any]

    def load(self, filename, expected_version_str=None, halt_on_error=True) -> dict:
        """Load a YAML file from disk.

        Args:
        ----
            filename: The file to load.
            expected_version_str: Version string to expect or None for no check.
            halt_on_error: Boolean which controls what happens if the file
                can't be loaded. (Not found, invalid format, etc. If True, MPF
                will raise an error and exit. If False, an empty config
                dictionary will be returned.

        Returns a dictionary of the settings from this YAML file.
        """
        if self.cache and filename in self.file_cache:
            return copy.deepcopy(self.file_cache[filename])

        config = dict()     # type: dict

        try:
            self.log.debug("Loading file: %s", filename)

            with open(filename, encoding='utf8') as f:
                if expected_version_str:
                    version_str = f.readline().strip()
                    if version_str != expected_version_str:
                        raise AssertionError("Version mismatch. Expected: {} Actual: {} Files: {}".format(
                            expected_version_str, version_str, filename))
                config = self.process(f)
                config = self.to_plain_dict(config)
        except MarkedYAMLError as e:
            mark = e.problem_mark
            msg = "YAML error found in file {}. Line {}, " \
                  "Position {}: {}".format(filename, mark.line + 1 if mark else None,
                                           mark.column + 1 if mark else None, e)

            if halt_on_error:
                raise ValueError(msg)

            self.log.warning(msg)
        except Exception as e:   # pylint: disable-msg=broad-except
            msg = "Error found in file {}: {}".format(filename, e)

            if halt_on_error:
                raise ValueError(msg)
            self.log.warning(msg)

        if self.cache and config:
            self.file_cache[filename] = copy.deepcopy(config)

        return config

    @staticmethod
    def to_plain_dict(data):
        """Recursively convert CommentedMap and CommentedSeq to Python dict and list respectively."""
        if isinstance(data, dict):
            return {key: YamlInterface.to_plain_dict(value) for key, value in data.items()}
        if isinstance(data, list):
            return [YamlInterface.to_plain_dict(item) for item in data]
        return data

    @staticmethod
    def process(data_string: Iterable[str]) -> dict:
        """Parse yaml from a string."""
        return _yaml.load(data_string)

    def save(self, filename: str, data: dict) -> None:   # pragma: no cover
        """Save config to yaml file."""
        with open(filename, 'w', encoding='utf8') as output_file:
            _yaml.default_flow_style = False
            _yaml.line_break = ''
            _yaml.dump(data, output_file)



# pylint: disable-msg=too-many-lines
"""PKONE Hardware interface.

Contains the hardware interface and drivers for the Penny K Pinball PKONE
platform hardware.
"""
import asyncio
from copy import deepcopy
from typing import Optional, Dict, List, Tuple, Set

from mpf.core.platform_batch_light_system import PlatformBatchLightSystem
from mpf.platforms.pkone.pkone_serial_communicator import PKONESerialCommunicator
from mpf.platforms.pkone.pkone_extension import PKONEExtensionBoard
from mpf.platforms.pkone.pkone_lightshow import PKONELightshowBoard
from mpf.platforms.pkone.pkone_switch import PKONESwitch, PKONESwitchNumber
from mpf.platforms.pkone.pkone_coil import PKONECoil, PKONECoilNumber
from mpf.platforms.pkone.pkone_servo import PKONEServo, PKONEServoNumber
from mpf.platforms.pkone.pkone_lights import PKONESimpleLED, PKONESimpleLEDNumber, PKONELEDChannel

from mpf.core.platform import SwitchPlatform, DriverPlatform, LightsPlatform, SwitchSettings, DriverSettings, \
    DriverConfig, SwitchConfig, RepulseSettings, ServoPlatform


# pylint: disable-msg=too-many-instance-attributes,too-many-public-methods
class PKONEHardwarePlatform(SwitchPlatform, DriverPlatform, LightsPlatform, ServoPlatform):

    """Platform class for the PKONE Nano hardware controller.

    Args:
    ----
        machine: The MachineController instance.
    """

    __slots__ = ["config", "serial_connections", "pkone_extensions", "pkone_lightshows", "_light_system",
                 "_watchdog_task", "hw_switch_data", "controller_connection", "pkone_commands"]

    def __init__(self, machine) -> None:
        """Initialize PKONE platform."""
        super().__init__(machine)
        self.controller_connection = None
        self.serial_connections = set()     # type: Set[PKONESerialCommunicator]
        self.pkone_extensions = {}          # type: Dict[int, PKONEExtensionBoard]
        self.pkone_lightshows = {}          # type: Dict[int, PKONELightshowBoard]
        self._light_system = None           # type: Optional[PlatformBatchLightSystem]
        self._watchdog_task = None
        self.hw_switch_data = dict()

        self.pkone_commands = {'PCN': lambda x, y: None,            # connected Nano processor
                               'PCB': lambda x, y: None,            # connected board
                               'PWD': lambda x, y: None,            # watchdog
                               'PWF': lambda x, y: None,            # watchdog stop
                               'PSA': self.receive_all_switches,    # all switch states
                               'PSW': self.receive_switch,          # switch state change
                               'PXX': self.receive_error,           # error
                               }

        # Set platform features. Each platform interface can change
        # these to notify the framework of the specific features it supports.
        self.features['max_pulse'] = 250
        self.features['tickless'] = True
        self.features['allow_empty_numbers'] = False

        self.config = self.machine.config_validator.validate_config("pkone", self.machine.config['pkone'])
        self._configure_device_logging_and_debug("PKONE", self.config)
        self.debug_log("Configuring PKONE hardware.")

    async def initialize(self):
        """Initialize connection to PKONE Nano hardware."""
        await self._connect_to_hardware()

        # Setup the batch light system
        self._light_system = PlatformBatchLightSystem(self.machine.clock, self._send_multiple_light_update,
                                                      self.machine.config['mpf']['default_light_hw_update_hz'],
                                                      64)

    def stop(self):
        """Stop platform and close connections."""
        if self._light_system:
            self._light_system.stop()
        if self.controller_connection:
            # send reset message to turn off all lights, disable all drivers, stop the watchdog process, etc.
            self.controller_connection.send('PRS')

        if self._watchdog_task:
            self._watchdog_task.cancel()
            self._watchdog_task = None

        # wait 100ms for the messages to be send
        self.machine.clock.loop.run_until_complete(asyncio.sleep(.1))

        if self.controller_connection:
            self.controller_connection.stop()
            self.controller_connection = None

        self.serial_connections = set()

    async def start(self):
        """Start listening for commands and schedule watchdog."""
        if self.config['watchdog']:
            # Configure the watchdog timeout interval and start it
            self.controller_connection.send('PWS{:04d}'.format(self.config['watchdog']))

            # Schedule the watchdog task to send at half the configured interval
            self._watchdog_task = self.machine.clock.schedule_interval(self._update_watchdog,
                                                                       self.config['watchdog'] / 2000)

        for connection in self.serial_connections:
            await connection.start_read_loop()

        self._initialize_led_hw_driver_alignment()
        self._light_system.start()

    def _update_watchdog(self):
        """Send Watchdog ping command."""
        self.controller_connection.send('PWD')

    def get_info_string(self):
        """Dump infos about boards."""
        if not self.serial_connections:
            return "No connection to any Penny K Pinball PKONE controller board."

        infos = "Penny K Pinball Hardware\n"
        infos += "------------------------\n"
        infos += " - Connected Controllers:\n"
        for connection in sorted(self.serial_connections, key=lambda x: x.port):
            infos += "   -> PKONE Nano - Port: {} at {} baud " \
                     "(firmware v{}, hardware rev {}).\n".format(connection.port,
                                                                 connection.baud,
                                                                 connection.remote_firmware,
                                                                 connection.remote_hardware_rev)

        infos += "\n - Extension boards:\n"
        for extension in self.pkone_extensions.values():
            infos += "   -> Address ID: {} (firmware v{}, hardware rev {})\n".format(extension.addr,
                                                                                     extension.firmware_version,
                                                                                     extension.hardware_rev)

        infos += "\n - Lightshow boards:\n"
        for lightshow in self.pkone_lightshows.values():
            infos += "   -> Address ID: {} ({} firmware v{}, " \
                     "hardware rev {})\n".format(lightshow.addr,
                                                 'RGBW' if lightshow.rgbw_firmware else 'RGB',
                                                 lightshow.firmware_version,
                                                 lightshow.hardware_rev)

        return infos

    async def _connect_to_hardware(self):
        """Connect to the port in the config."""
        comm = PKONESerialCommunicator(platform=self, port=self.config['port'], baud=self.config['baud'])
        await comm.connect()
        self.serial_connections.add(comm)

    def register_extension_board(self, board: PKONEExtensionBoard):
        """Register an Extension board."""
        if board.addr in self.pkone_extensions or board.addr in self.pkone_lightshows:
            raise AssertionError("Duplicate address id: a board has already been "
                                 "registered at address {}".format(board.addr))

        if board.addr not in range(8):
            raise AssertionError("Address out of range: Extension board address id must be between 0 and 7")

        self.pkone_extensions[board.addr] = board

    def register_lightshow_board(self, board: PKONELightshowBoard):
        """Register a Lightshow board."""
        if board.addr in self.pkone_extensions or board.addr in self.pkone_lightshows:
            raise AssertionError("Duplicate address id: a board has already been "
                                 "registered at address {}".format(board.addr))

        if board.addr not in range(4):
            raise AssertionError("Address out of range: Lightshow board address id must be between 0 and 3")

        self.pkone_lightshows[board.addr] = board

    def process_received_message(self, msg: str):
        """Send an incoming message from the PKONE controller to the proper method for servicing.

        Args:
        ----
            msg: messaged which was received
        """
        assert self.log is not None
        cmd = msg[0:3]
        payload = msg[3:].replace('E', '')

        # Can't use try since it swallows too many errors for now
        if cmd in self.pkone_commands:
            self.pkone_commands[cmd](payload)
        else:   # pragma: no cover
            self.log.warning("Received unknown serial command %s.", msg)

    def receive_error(self, msg):
        """Receive an error message from the controller."""
        self.log.error("Received an error message from the controller: %s", msg)

    def _parse_coil_number(self, number: str) -> PKONECoilNumber:
        try:
            board_id_str, coil_num_str = number.split("-")
        except ValueError:
            raise AssertionError("Invalid coil number {}".format(number))

        board_id = int(board_id_str)
        coil_num = int(coil_num_str)

        if board_id not in self.pkone_extensions:
            raise AssertionError("PKONE Extension {} does not exist for coil {}".format(board_id, number))

        if coil_num == 0:
            raise AssertionError("PKONE coil numbering begins with 1. Coil: {}".format(number))

        coil_count = self.pkone_extensions[board_id].coil_count
        if coil_count < coil_num or coil_num < 1:
            raise AssertionError(
                "PKONE Extension {board_id} only has {coil_count} coils "
                "({first_coil} - {last_coil}). Coil: {number}".format(
                    board_id=board_id, coil_count=coil_count, first_coil=1, last_coil=coil_count, number=number))

        return PKONECoilNumber(board_id, coil_num)

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict) -> PKONECoil:
        """Configure a coil/driver.

        Args:
        ----
            config: Coil/driver config.
            number: Number of this coil/driver.
            platform_settings: Platform specific settings.

        Returns: Coil/driver object
        """
        # dont modify the config. make a copy
        platform_settings = deepcopy(platform_settings)

        if not self.controller_connection:
            raise AssertionError('A request was made to configure a PKONE coil, but no '
                                 'connection to a PKONE controller is available')

        if not number:
            raise AssertionError("Coil number is required")

        coil_number = self._parse_coil_number(str(number))
        return PKONECoil(config, self, coil_number, platform_settings)

    @staticmethod
    def _check_coil_switch_combination(coil: DriverSettings, switch: SwitchSettings):
        """Check to see if the coil/switch combination is legal for hardware rules."""
        # coil and switch must be on the same extension board (same board address id)
        if switch.hw_switch.number.board_address_id != coil.hw_driver.number.board_address_id:
            raise AssertionError("Coil {} and switch {} are on different boards. Cannot apply hardware rule!".format(
                coil.hw_driver.number, switch.hw_switch.number))

    def clear_hw_rule(self, switch: SwitchSettings, coil: DriverSettings):
        """Clear a hardware rule.

        This is used if you want to remove the linkage between a switch and
        some coil activity. For example, if you wanted to disable your
        flippers (so that a player pushing the flipper buttons wouldn't cause
        the flippers to flip), you'd call this method with your flipper button
        as the *sw_num*.

        Args:
        ----
            switch: The switch whose rule you want to clear.
            coil: The coil whose rule you want to clear.
        """
        self.debug_log("Clearing Hardware Rule for coil: %s, switch: %s",
                       coil.hw_driver.number, switch.hw_switch.number)
        driver = coil.hw_driver
        driver.clear_hardware_rule()

    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit rule on driver.

        Pulses a driver when a switch is hit. When the switch is released the pulse continues. Typically used for
        autofire coils such as pop bumpers.
        """
        self._check_coil_switch_combination(coil, enable_switch)
        driver = coil.hw_driver
        driver.set_hardware_rule(1, enable_switch, None, 0, coil.pulse_settings, None)

    def set_delayed_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings, delay_ms: int):
        """Set pulse on hit and release rule to driver.

        When a switch is hit and a certain delay passed it pulses a driver.
        When the switch is released the pulse continues.
        Typically used for kickbacks.
        """
        self._check_coil_switch_combination(coil, enable_switch)
        driver = coil.hw_driver
        driver.set_hardware_rule(2, enable_switch, None, delay_ms, coil.pulse_settings, None)

    def set_pulse_on_hit_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and release rule to driver.

        Pulses a driver when a switch is hit. When the switch is released the pulse is canceled. Typically used on
        the main coil for dual coil flippers without eos switch.
        """
        self._check_coil_switch_combination(coil, enable_switch)
        driver = coil.hw_driver
        driver.set_hardware_rule(3, enable_switch, None, 0, coil.pulse_settings, coil.hold_settings)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and enable and release rule on driver.

        Pulses a driver when a switch is hit. Then enables the driver (may be with pwm). When the switch is released
        the pulse is canceled and the driver gets disabled. Typically used for single coil flippers.
        """
        self._check_coil_switch_combination(coil, enable_switch)
        driver = coil.hw_driver
        driver.set_hardware_rule(4, enable_switch, None, 0, coil.pulse_settings, coil.hold_settings)

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings, eos_switch: SwitchSettings,
                                                      coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver.

        Pulses a driver when a switch is hit. When the switch is released
        the pulse is canceled and the driver gets disabled. When the eos_switch is hit the pulse is canceled
        and the driver becomes disabled. Typically used on the main coil for dual-wound coil flippers with eos switch.
        """
        del repulse_settings
        self._check_coil_switch_combination(coil, enable_switch)
        self._check_coil_switch_combination(coil, eos_switch)
        driver = coil.hw_driver
        driver.set_hardware_rule(5, enable_switch, eos_switch, 0, coil.pulse_settings, coil.hold_settings)

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver.

        Pulses a driver when a switch is hit. Then enables the driver (may be with pwm). When the switch is released
        the pulse is canceled and the driver becomes disabled. When the eos_switch is hit the pulse is canceled
        and the driver becomes enabled (likely with PWM).
        Typically used on the coil for single-wound coil flippers with eos switch.
        """
        raise AssertionError("Single-wound coils with EOS are not implemented in PKONE hardware.")

    def _parse_servo_number(self, number: str) -> PKONEServoNumber:
        try:
            board_id_str, servo_num_str = number.split("-")
        except ValueError:
            raise AssertionError("Invalid servo number {}".format(number))

        board_id = int(board_id_str)
        servo_num = int(servo_num_str)

        if board_id not in self.pkone_extensions:
            raise AssertionError("PKONE Extension {} does not exist for servo {}".format(board_id, number))

        # Servos are numbered in sequence immediately after the highest coil number
        driver_count = self.pkone_extensions[board_id].coil_count
        servo_count = self.pkone_extensions[board_id].servo_count
        if servo_num <= driver_count or servo_num > driver_count + servo_count:
            raise AssertionError("PKONE Extension {} supports {} servos ({} - {}). "
                                 "Servo: {} is not a valid number.".format(
                                     board_id, servo_count, driver_count + 1, driver_count + servo_count, number))

        return PKONEServoNumber(board_id, servo_num)

    async def configure_servo(self, number: str, config: dict) -> PKONEServo:
        """Configure a servo."""
        del config
        servo_number = self._parse_servo_number(str(number))
        return PKONEServo(servo_number, self.controller_connection)

    @classmethod
    def get_coil_config_section(cls):
        """Return coil config section."""
        return "pkone_coils"

    def _parse_switch_number(self, number: str) -> PKONESwitchNumber:
        try:
            board_id_str, switch_num_str = number.split("-")
        except ValueError:
            raise AssertionError("Invalid switch number {}".format(number))

        board_id = int(board_id_str)
        switch_num = int(switch_num_str)

        if board_id not in self.pkone_extensions:
            raise AssertionError("PKONE Extension {} does not exist for switch {}".format(board_id, number))

        if switch_num == 0:
            raise AssertionError("PKONE switch numbering begins with 1. Switch: {}".format(number))

        if self.pkone_extensions[board_id].switch_count < switch_num:
            raise AssertionError("PKONE Extension {} only has {} switches. Switch: {}".format(
                board_id, self.pkone_extensions[board_id].switch_count, number))

        return PKONESwitchNumber(board_id, switch_num)

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict) -> PKONESwitch:
        """Configure the switch object for a PKONE controller.

        Args:
        ----
            number: Number of this switch.
            config: Switch config.
            platform_config: Platform specific settings.

        Returns: Switch object.
        """
        del platform_config
        if not number:
            raise AssertionError("Switch requires a number")

        if not self.controller_connection:
            raise AssertionError("A request was made to configure a PKONE switch, but no "
                                 "connection to PKONE controller is available")

        try:
            switch_number = self._parse_switch_number(number)
        except ValueError:
            raise AssertionError("Could not parse switch number {}/{}. Seems "
                                 "to be not a valid switch number for the"
                                 "PKONE platform.".format(config.name, number))

        self.debug_log("PKONE Switch: %s (%s)", number, config.name)
        return PKONESwitch(config, switch_number, self)

    async def get_hw_switch_states(self) -> Dict[str, bool]:
        """Return hardware states."""
        return self.hw_switch_data

    def receive_all_switches(self, msg):
        """Process the all switch states message."""
        # The PSA message contains the following information:
        # [PSA opcode] + [[board address id] + 0 or 1 for each switch on the board] + E
        self.debug_log("Received all switch states (PSA): %s", msg)

        # the message payload is delimited with an 'X' character for the switches on each board
        # The first character is the board address ID
        board_address_id = int(msg[0])
        switch_states = msg[1:]

        # There is one character for each switch on the board (1 = active, 0 = inactive)
        # Loop over each character and map the state to the appropriate switch number
        for index, state in enumerate(switch_states):
            self.hw_switch_data[PKONESwitchNumber(board_address_id=board_address_id,
                                                  switch_number=index + 1)] = int(state)

    def receive_switch(self, msg):
        """Process a single switch state change."""
        # The PSW message contains the following information:
        # [PSW opcode] + [board address id] + switch number + switch state (0 or 1) + E
        self.debug_log("Received switch state change (PSW): %s", msg)
        switch_number = PKONESwitchNumber(int(msg[0]), int(msg[1:3]))
        switch_state = int(msg[-1])
        self.machine.switch_controller.process_switch_by_num(state=switch_state,
                                                             num=switch_number,
                                                             platform=self)

    # pylint: disable-msg=too-many-locals
    async def _send_multiple_light_update(self, sequential_brightness_list: List[Tuple[PKONELEDChannel,
                                                                                       float, int]]):
        # determine how many channels are to be updated and the common fade time
        first_channel, _, common_fade_ms = sequential_brightness_list[0]
        start_index = first_channel.index

        # get the lightshow board that will be sent the command (need to know if rgb or rgbw board)
        lightshow = self.pkone_lightshows[first_channel.board_address_id]
        if lightshow.rgbw_firmware:
            cmd_opcode = "PWB"
            channel_grouping = 4
        else:
            cmd_opcode = "PLB"
            channel_grouping = 3

        # determine if first and last batch channels are properly aligned to internal LED boundaries
        first_channel_alignment_offset = first_channel.index % channel_grouping
        last_channel_alignment_offset = (first_channel.index + len(sequential_brightness_list)) % channel_grouping

        # Note: software fading will automatically be used when batch channels
        # are not aligned to hardware LED boundaries

        if first_channel_alignment_offset > 0:
            # the first channel does not align with internal hardware boundary, need to retrieve other
            # channels for the first light
            current_time = self._light_system.clock.get_time()
            channel = first_channel
            previous_channel = None
            for _ in range(first_channel_alignment_offset):
                if channel:
                    previous_channel = lightshow.get_channel_hw_driver(first_channel.group,
                                                                       channel.get_predecessor_number())
                if previous_channel:
                    brightness, fade_ms, _ = previous_channel.get_fade_and_brightness(current_time)
                    channel = previous_channel
                else:
                    brightness = 0.0
                    fade_ms = 0
                    channel = None

                sequential_brightness_list.insert(0, (channel, brightness, fade_ms))
                start_index = start_index - 1

        if last_channel_alignment_offset > 0:
            current_time = self._light_system.clock.get_time()
            channel = sequential_brightness_list[-1][0]
            next_channel = None
            for _ in range(channel_grouping - last_channel_alignment_offset):
                if channel:
                    next_channel = lightshow.get_channel_hw_driver(first_channel.group,
                                                                   channel.get_successor_number())
                if next_channel:
                    brightness, fade_ms, _ = next_channel.get_fade_and_brightness(current_time)
                    channel = next_channel
                else:
                    brightness = 0.0
                    fade_ms = 0
                    channel = None

                sequential_brightness_list.append((channel, brightness, fade_ms))

        assert start_index % channel_grouping == 0
        assert len(sequential_brightness_list) % channel_grouping == 0

        # generate batch update command using brightness values (3 digit int values)
        # Note: fade time is in 10ms units
        cmd = "{}{}{}{:02d}{:02d}{:04d}{}".format(cmd_opcode,
                                                  first_channel.board_address_id,
                                                  first_channel.group,
                                                  start_index // channel_grouping + 1,
                                                  len(sequential_brightness_list) // channel_grouping,
                                                  int(common_fade_ms / 10),
                                                  "".join("%03d" % (b[1] * 255) for b in sequential_brightness_list))
        self.controller_connection.send(cmd)

    def configure_light(self, number, subtype, config, platform_settings):
        """Configure light in platform."""
        del platform_settings

        if not self.controller_connection:
            raise AssertionError("A request was made to configure a PKONE light, but no "
                                 "connection to PKONE controller is available")

        if subtype == "simple":
            # simple LEDs use the format <board_address_id> - <led> (simple LEDs only have 1 channel)
            board_address_id, index = number.split('-')
            return PKONESimpleLED(PKONESimpleLEDNumber(int(board_address_id), int(index)),
                                  self.controller_connection.send, self)

        if not subtype or subtype == "led":
            board_address_id, group, index = number.split("-")
            led_channel = PKONELEDChannel(board_address_id, group, index, config, self._light_system)
            lightshow = self.pkone_lightshows[int(board_address_id)]
            lightshow.add_channel_hw_driver(int(group), led_channel)
            self._light_system.mark_dirty(led_channel)
            return led_channel

        raise AssertionError("Unknown subtype {}".format(subtype))

    def _led_is_hardware_aligned(self, led_name) -> bool:
        """Determine whether the specified LED is hardware aligned."""
        light = self.machine.lights[led_name]
        hw_numbers = light.get_hw_numbers()
        board_address_id, _, index = hw_numbers[0].split("-")
        lightshow = self.pkone_lightshows[int(board_address_id)]
        if lightshow.rgbw_firmware:
            channel_grouping = 4
        else:
            channel_grouping = 3
        if len(hw_numbers) != channel_grouping:
            return False
        return int(index) % channel_grouping == 0

    def _initialize_led_hw_driver_alignment(self):
        """Set hardware aligned flag for all led hardware driver channels."""
        for _, lightshow in self.pkone_lightshows.items():
            for channel in lightshow.get_all_channel_hw_drivers():
                channel.set_hardware_aligned(self._led_is_hardware_aligned(channel.config.name))

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse light channels from number string."""
        if subtype == "simple":
            # simple LEDs use the format <board_address_id> - <led> (simple LEDs only have 1 channel)
            board_address_id, index = number.split('-')
            return [
                {
                    "number": "{}-{}".format(board_address_id, index)
                }
            ]

        if not subtype or subtype == "led":
            # Normal LED number format: <board_address_id> - <group> - <led>
            board_address_id, group, number_str = str(number).split('-')
            index = int(number_str)

            # Determine if there are 3 or 4 channels depending upon firmware on board
            if self.pkone_lightshows[board_address_id].rgbw_firmware:
                # rgbw uses 4 channels per led
                return [
                    {
                        "number": "{}-{}-{}".format(board_address_id, group, (index - 1) * 4)
                    },
                    {
                        "number": "{}-{}-{}".format(board_address_id, group, (index - 1) * 4 + 1)
                    },
                    {
                        "number": "{}-{}-{}".format(board_address_id, group, (index - 1) * 4 + 2)
                    },
                    {
                        "number": "{}-{}-{}".format(board_address_id, group, (index - 1) * 4 + 3)
                    },
                ]

            # rgb uses 3 channels per led
            return [
                {
                    "number": "{}-{}-{}".format(board_address_id, group, (index - 1) * 3)
                },
                {
                    "number": "{}-{}-{}".format(board_address_id, group, (index - 1) * 3 + 1)
                },
                {
                    "number": "{}-{}-{}".format(board_address_id, group, (index - 1) * 3 + 2)
                },
            ]

        raise AssertionError("Unknown light subtype {}".format(subtype))



"""Event Config Player."""
from copy import deepcopy

from mpf.core.placeholder_manager import TextTemplate

from mpf.config_players.flat_config_player import FlatConfigPlayer
from mpf.core.delays import DelayManager
from mpf.core.utility_functions import Util


class EventPlayer(FlatConfigPlayer):

    """Posts events based on config."""

    config_file_section = 'event_player'
    show_section = 'events'

    __slots__ = ["delay"]

    def __init__(self, machine):
        """Initialize EventPlayer."""
        super().__init__(machine)
        self.delay = DelayManager(self.machine)

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Post (delayed) events."""
        for event, event_configs in settings.items():
            for s in event_configs:
                if s["condition"] and not s["condition"].evaluate(kwargs):
                    continue

                if s["number"] is not None:
                    self.delay.add(callback=self._post_event, ms=s["number"],
                                   event=event, priority=s["priority"], params=s["params"], **kwargs)
                else:
                    self._post_event(event, s["priority"], s["params"], **kwargs)

    # pylint: disable-msg=too-many-arguments
    def handle_subscription_change(self, value, settings, priority, context, key):
        """Handle subscriptions."""
        instance_dict = self._get_instance_dict(context)
        previous_value = instance_dict.get(key, None)
        if previous_value == value:
            return
        instance_dict[key] = value

        if not value:
            return

        for event, event_configs in settings.items():
            for s in event_configs:
                if s["condition"] and not s["condition"].evaluate({}):
                    continue

                if s["number"] is not None:
                    self.delay.add(callback=self._post_event, ms=s["number"],
                                   event=event, priority=s["priority"], params=s["params"])
                else:
                    self._post_event(event, s["priority"], s["params"])

    def _post_event(self, event, priority, params, **kwargs):
        if "(" in event:
            # TODO: move this to parsing time
            event_name_placeholder = TextTemplate(self.machine, event.replace("(", "{").replace(")", "}"))
            event = event_name_placeholder.evaluate(kwargs)
        for key, param in params.items():
            if isinstance(param, dict):
                params = deepcopy(params)
                # TODO: move this to parsing time
                params[key] = self._evaluate_event_param(param, kwargs)
        self.machine.events.post(event, priority=priority, **params)

    def validate_config_entry(self, settings, name):
        """Validate player and expand placeholders."""
        config = super().validate_config_entry(settings, name)
        final_config = {}
        for event, s in config.items():
            if "(" in event:
                if event not in final_config:
                    final_config[event] = []
                final_config[event].append({
                    "condition": None,
                    "number": None,
                    "priority": s["priority"],
                    "params": {k: v for k, v in s.items() if k != "priority"}
                })
            else:
                var = self._parse_and_validate_conditional(event, name)
                if var.name not in final_config:
                    final_config[var.name] = []
                final_config[var.name].append({
                    "condition": var.condition,
                    "number": Util.string_to_ms(var.number) if var.number else None,
                    "priority": s["priority"],
                    "params": {k: v for k, v in s.items() if k != "priority"}
                })
        return final_config

    def _evaluate_event_param(self, param, kwargs):
        if param.get("type") == "float":
            placeholder = self.machine.placeholder_manager.build_float_template(param["value"])
        elif param.get("type") == "int":
            placeholder = self.machine.placeholder_manager.build_int_template(param["value"])
        elif param.get("type") == "bool":
            placeholder = self.machine.placeholder_manager.build_bool_template(param["value"])
        else:
            placeholder = self.machine.placeholder_manager.build_string_template(param["value"])
        return placeholder.evaluate(kwargs)

    def get_list_config(self, value):
        """Parse list."""
        result = {}
        for event in value:
            result[event] = {}
        return result

    def get_express_config(self, value):
        """Parse short config."""
        return self.get_list_config(Util.string_to_event_list(value))



"""A switch connected to a FAST controller."""
import logging
from copy import copy
from dataclasses import dataclass

from mpf.core.utility_functions import Util

MYPY = False


@dataclass
class FastSwitchConfig:

    """Configuration class object for a FAST switch configuration."""

    number: str          # '00'-'68'
    mode: str            # '00'-'02'
    debounce_close: str  # '00'-'FF'
    debounce_open: str   # '00'-'FF'


class FASTSwitch:

    """A switch connected to a FAST controller."""

    __slots__ = ["log", "communicator", "number", "hw_number", "platform", "baseline_switch_config",
                 "current_hw_config"]

    def __init__(self, communicator, hw_number, ) -> None:
        """Initialize switch."""
        self.log = logging.getLogger('FAST Switch')
        self.communicator = communicator
        self.number = hw_number  # must be int to work with the rest of MPF
        self.hw_number = Util.int_to_hex_string(hw_number)  # hex version the FAST hw actually uses
        self.platform = communicator.platform  # Needed by the SwitchController

        self.baseline_switch_config = FastSwitchConfig(number=self.hw_number, mode='00', debounce_close='00',
                                                       debounce_open='00')
        self.current_hw_config = self.baseline_switch_config

    def set_initial_config(self, mpf_config, platform_settings):
        """Takes the mpf_config and platform_settings and converts them to the FAST format.

        Sets that to the current config and establishes it as the baseline. This is only called
        when MPF is starting up.
        """
        self.current_hw_config = self.convert_mpf_config_to_fast(mpf_config, platform_settings)
        self.baseline_switch_config = copy(self.current_hw_config)

    def convert_mpf_config_to_fast(self, mpf_config, platform_settings) -> FastSwitchConfig:
        """Converts the MPF switch config to the FAST format."""
        debounce_close, debounce_open = self.reconcile_debounce(mpf_config, platform_settings)

        if mpf_config.invert:
            mode = '02'
            debounce_close, debounce_open = debounce_open, debounce_close
        else:
            mode = '01'

        return FastSwitchConfig(number=self.hw_number, mode=mode, debounce_close=debounce_close,
                                debounce_open=debounce_open)

    def send_config_to_switch(self):
        """Sends the current config to the switch. e.g. actually updates the physical switch."""
        msg = (f'{self.communicator.SWITCH_CMD}:{self.current_hw_config.number},'
               f'{self.current_hw_config.mode},{self.current_hw_config.debounce_close},'
               f'{self.current_hw_config.debounce_open}')

        self.communicator.send_with_confirmation(msg, f'{self.communicator.SWITCH_CMD}:P')

    def get_current_config(self):
        """Return the current config of this switch."""
        return (f'{self.communicator.SWITCH_CMD}:{self.current_hw_config.number},'
                f'{self.current_hw_config.mode},{self.current_hw_config.debounce_close},'
                f'{self.current_hw_config.debounce_open}')

    def get_board_name(self):
        """Return the board of this switch."""
        switch_index = 0
        for board_obj in self.communicator.platform.io_boards.values():
            if switch_index <= self.number < switch_index + board_obj.switch_count:
                return f"FAST Board {str(board_obj.node_id)}"
            switch_index += board_obj.switch_count

        # fall back if not found
        return "FAST Unknown Board"

    def reconcile_debounce(self, config, platform_settings):
        """Looks at all the possible debounce settings.

        Checks the switch, the platform, and FAST overrides,
        and returns the final debounce settings for the switch.

        Returns a tuple of (debounce_open, debounce_close) in hex string format.
        """
        # Set the debounce from the generic true/false first, then override
        # if the platform settings have specific values
        if config.debounce in ['normal', 'auto']:
            debounce_open = self.communicator.config['default_normal_debounce_open']
            debounce_close = self.communicator.config['default_normal_debounce_close']
        else:
            debounce_open = self.communicator.config['default_quick_debounce_open']
            debounce_close = self.communicator.config['default_quick_debounce_close']

        if 'debounce_open' in platform_settings and platform_settings['debounce_open'] is not None:
            debounce_open = platform_settings['debounce_open']

        if 'debounce_close' in platform_settings and platform_settings['debounce_close'] is not None:
            debounce_close = platform_settings['debounce_close']

        return Util.int_to_hex_string(debounce_close), Util.int_to_hex_string(debounce_open)



"""Retro Platform Serial Communicator."""
# mpf/platforms/fast/communicators/net_retro.py

from mpf.platforms.fast.communicators.net_neuron import FastNetNeuronCommunicator


class FastNetRetroCommunicator(FastNetNeuronCommunicator):

    """Serial communicator class for FAST Retro platform."""

    MAX_SWITCHES = 96  # prev version was 128
    MAX_DRIVERS = 48   # prev version was 64
    IGNORED_MESSAGES = ['WD:P', 'TL:P', 'L1:P', 'GI:P']

    async def query_io_boards(self):
        """Query the board for switches and drivers.

        There are no external I/O boards on a Retro Controller, so
        instead we create one to represent the onboard switches & drivers.
        """
        self.io_loop = ['retro']  # Manually create the internal I/O board since it's not in the config
        self.config['io_loop']['retro'] = {'model': 'FP-RETRO-I/O',
                                           'order': 0,
                                           }

        # Simplest way is just to fake a response to the NN: query
        # NN:<NODE_ID>,<NODE_NAME>,<NODE_FIRMWARE>,<DRIVER_COUNT>,<SWITCH_COUNT>,<EXTRADATA_IN>,<EXTRADATA_OUT>
        self._process_nn(
            f'00,FP-RETRO-I/O,{self.remote_firmware},{self.MAX_DRIVERS:02X},{self.MAX_SWITCHES:02X},0,0,0,0,0,0'
        )



# mpf.tests.test_Fast_Seg (Segment Displays)

from mpf.tests.test_Fast import TestFastBase
from mpf.tests.MpfTestCase import test_config

class TestFastSeg(TestFastBase):
    """Tests the FAST Audio Interface boards."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.serial_connections_to_mock = ['seg']

    def get_config_file(self):
        return 'seg.yaml'

    def create_expected_commands(self):
        # self.serial_connections['seg'].expected_commands = {
        #         'PA:00,       ': '',}
        pass

    def test_seg(self):
        # This was migrated from MPF 0.56 and has not been tested with real hardware.
        # There's a chance it doesn't actually work, but should be simple to fix.

        self.assertEqual(5, len(self.machine.segment_displays))

        seg_addresses = list()

        # Make sure the MPF segment interface is picking up the digits and colors
        for seg in self.machine.segment_displays:
            for digit_color in seg.colors:
                self.assertEqual('ffffff', digit_color.hex)
            for digit_text in seg.text:
                self.assertEqual(' ', digit_text)

            num_digits = seg.size
            fast_address = seg.hw_display.hex_id
            seg_addresses.append(fast_address)

            # make sure the commands sent via FSP match the lengths of digits
            self.assertEqual(len(self.seg_cpu.seg_digits[fast_address]), num_digits)
            self.assertEqual(len(self.seg_cpu.seg_colors[fast_address]), num_digits)

        # Make sure the serial commands went out
        for display, colors in self.seg_cpu.seg_colors.items():
            for digit in colors:
                self.assertEqual('FFFFFF', digit)